from pychess.System import conf
from pychess.Utils.const import FAN_PIECES, BLACK, WHITE

__label__ = _("Html Diagram")
__ending__ = "html"
__append__ = False

SIZE = 40
BORDER_SIZE = SIZE // 4 + SIZE // 8
FILL = SIZE - BORDER_SIZE
FONT_SIZE = SIZE - 4

#    font-family: "ChessMedium";

style = """
.chessboard {{
    width: {}px;
    height: {}px;
    font-family: "DejaVu Serif", "DejaVu", serif;
    line-height: {}px;
}}
.black {{
    float: left;
    width: {}px;
    height: {}px;
    background-color: #999;
    font-size:{}px;
    text-align:center;
    display: table-cell;
    vertical-align:middle;
}}
.white {{
    float: left;
    width: {}px;
    height: {}px;
    background-color: #fff;
    font-size:{}px;
    text-align:center;
    display: table-cell;
    vertical-align:middle;
}}
.fill-top {{
    float: left;
    width: {}px;
    height: {}px;
    color: #ffff;
    display: table-cell;
}}
.top-corner {{
    float: left;
    width: {}px;
    height: {}px;
    background-color: #333;
    display: table-cell;
}}
.top {{
    float: left;
    width: {}px;
    height: {}px;
    background-color: #333;
    display: table-cell;
}}
.fill-side {{
    float: left;
    width: {}px;
    height: {}px;
    color: #ffff;
    display: table-cell;
}}
.side {{
    float: left;
    width: {}px;
    height: {}px;
    color: #ffff;
    background-color: #333;
    font-size: {}px;
    text-align:center;
    display: table-cell;
    vertical-align:middle;
}}
.bottom-corner {{
    float: left;
    width: {}px;
    height: {}px;
    background-color: #333;
    display: table-cell;
}}
.bottom {{
    float: left;
    width: {}px;
    height: {}px;
    color: #ffff;
    background-color: #333;
    font-size: {}px;
    line-height: {}px;
    text-align: center;
    display: table-cell;
}}
""".format(
    SIZE * 10,
    SIZE * 10,
    SIZE,  # chessboard
    SIZE,
    SIZE,
    FONT_SIZE,  # black
    SIZE,
    SIZE,
    FONT_SIZE,  # white
    FILL,
    BORDER_SIZE,  # fill-top
    BORDER_SIZE,
    BORDER_SIZE,  # top-corner
    SIZE,
    BORDER_SIZE,  # top
    FILL,
    SIZE,  # fill-side
    BORDER_SIZE,
    SIZE,
    BORDER_SIZE,  # side
    BORDER_SIZE,
    BORDER_SIZE,  # bottom-corner
    SIZE,
    BORDER_SIZE,
    BORDER_SIZE,
    BORDER_SIZE,  # bottom
)


def save(file, model, position=None, flip=False):
    """Export the current position into a .html file using html+css"""

    print(
        "<html><head><meta http-equiv='Content-Type' content='text/html;charset=UTF-8'>",
        file=file,
    )
    #    print('<link rel="stylesheet" media="screen" href="https://fontlibrary.org/face/chess" type="text/css"/>', file=file)
    print("<style type='text/css'>%s" % style, file=file)
    print("</style></head><body><div class='chessboard'>", file=file)

    show_cords = conf.get("showCords")
    cords_side = "12345678" if flip else "87654321"
    cords_bottom = "HGFEDCBA" if flip else "ABCDEFGH"

    data = model.boards[position].data[:]

    board = ""
    # header
    if show_cords:
        board += "<div class='fill-top'></div>"
        board += "<div class='top-corner'></div>"
        for cord in range(8):
            board += "<div class='top'></div>"
        board += "<div class='top-corner'></div>"
        board += "<div class='fill-top'></div>"
    for j, row in enumerate(data if flip else reversed(data)):
        if show_cords:
            board += "<div class='fill-side'></div>"
            board += "<div class='side'>%s</div>" % cords_side[j]
        for i in range(8):
            if j % 2 == 0:
                color = "white" if i % 2 == 0 else "black"
            else:
                color = "white" if i % 2 == 1 else "black"

            piece = row.get(i)
            if piece is not None:
                if piece.color == BLACK:
                    piece_fan = FAN_PIECES[BLACK][piece.piece]
                else:
                    piece_fan = FAN_PIECES[WHITE][piece.piece]
                board += f"<div class='{color}'>{piece_fan}</div>"
            else:
                board += "<div class='%s'></div>" % color
        if show_cords:
            board += "<div class='side'></div>"
            board += "<div class='fill-side'></div>"
        board += "\n"

    if show_cords:
        board += "<div class='fill-top'></div>"
        board += "<div class='bottom-corner'></div>"
        for cord in cords_bottom:
            board += "<div class='bottom'>%s</div>" % cord
        board += "<div class='bottom-corner'></div>"
        board += "<div class='fill-top'></div>"

    print(board, file=file)

    print("</div></body></html>", file=file)

    file.close()


if __name__ == "__main__":
    from pychess.Utils.GameModel import GameModel

    model = GameModel()
    with open("/home/tamas/board.html", "w") as fi:
        save(fi, model, position=0, flip=True)



from gi.repository import Gtk, Gdk, GObject

from pychess.widgets import mainwindow


class ImageMenu(Gtk.EventBox):
    def __init__(self, image, child):
        GObject.GObject.__init__(self)
        self.add(image)

        self.subwindow = Gtk.Window()
        self.subwindow.set_transient_for(mainwindow())
        self.subwindow.set_decorated(False)
        self.subwindow.set_resizable(False)
        self.subwindow.set_type_hint(Gdk.WindowTypeHint.DIALOG)
        self.subwindow.add(child)
        self.subwindow.connect_after("draw", self.__sub_onExpose)
        self.subwindow.connect("button_press_event", self.__sub_onPress)
        # self.subwindow.connect("motion_notify_event", self.__sub_onMotion)
        # self.subwindow.connect("leave_notify_event", self.__sub_onMotion)
        # self.subwindow.connect("delete-event", self.__sub_onDelete)
        self.subwindow.connect("focus-out-event", self.__sub_onFocusOut)
        child.show_all()

        self.setOpen(False)
        self.connect("button_press_event", self.__onPress)

    def setOpen(self, isopen):
        self.isopen = isopen

        if isopen:
            topwindow = self.get_parent()
            while not isinstance(topwindow, Gtk.Window):
                topwindow = topwindow.get_parent()
            x_loc, y_loc = topwindow.get_window().get_position()
            x_loc += self.get_allocation().x + self.get_allocation().width
            y_loc += self.get_allocation().y
            self.subwindow.move(x_loc, y_loc)

        self.subwindow.props.visible = isopen
        self.set_state(self.isopen and Gtk.StateType.SELECTED or Gtk.StateType.NORMAL)

    def __onPress(self, self_, event):
        if event.button == 1 and event.type == Gdk.EventType.BUTTON_PRESS:
            self.setOpen(not self.isopen)

    def __sub_setGrabbed(self, grabbed):
        if grabbed and not Gdk.pointer_is_grabbed():
            Gdk.pointer_grab(
                self.subwindow.get_window(),
                True,
                Gdk.EventMask.LEAVE_NOTIFY_MASK
                | Gdk.EventMask.POINTER_MOTION_MASK
                | Gdk.EventMask.BUTTON_PRESS_MASK,
                None,
                None,
                Gdk.CURRENT_TIME,
            )
            Gdk.keyboard_grab(self.subwindow.get_window(), True, Gdk.CURRENT_TIME)
        elif Gdk.pointer_is_grabbed():
            Gdk.pointer_ungrab(Gdk.CURRENT_TIME)
            Gdk.keyboard_ungrab(Gdk.CURRENT_TIME)

    def __sub_onMotion(self, subwindow, event):
        allocation = subwindow.get_allocation()
        self.__sub_setGrabbed(
            not (0 <= event.x < allocation.width and 0 <= event.y < allocation.height)
        )

    def __sub_onPress(self, subwindow, event):
        allocation = subwindow.get_allocation()
        if not (0 <= event.x < allocation.width and 0 <= event.y < allocation.height):
            Gdk.pointer_ungrab(event.time)
            self.setOpen(False)

    def __sub_onExpose(self, subwindow, ctx):
        allocation = subwindow.get_allocation()
        context = subwindow.get_window().cairo_create()
        context.set_line_width(2)
        context.rectangle(
            allocation.x, allocation.y, allocation.width, allocation.height
        )
        style_ctxt = self.get_style_context()
        color = style_ctxt.lookup_color("p_dark_color")[1]
        red, green, blue, alpha = color.red, color.green, color.blue, color.alpha
        context.set_source_rgba(red, green, blue, alpha)
        context.stroke()
        # self.__sub_setGrabbed(self.isopen)

    def __sub_onDelete(self, subwindow, event):
        self.setOpen(False)
        return True

    def __sub_onFocusOut(self, subwindow, event):
        self.setOpen(False)


def switchWithImage(image, dialog):
    parent = image.get_parent()
    parent.remove(image)
    imageMenu = ImageMenu(image, dialog)
    parent.add(imageMenu)
    imageMenu.show()


if __name__ == "__main__":
    win = Gtk.Window()
    vbox = Gtk.VBox()
    vbox.add(Gtk.Label(label="Her er der en kat"))
    image = Gtk.Image.new_from_icon_name("gtk-properties", Gtk.IconSize.BUTTON)
    vbox.add(image)
    vbox.add(Gtk.Label(label="Her er der ikke en kat"))
    win.add(vbox)

    table = Gtk.Table(2, 2)
    table.attach(Gtk.Label(label="Minutes:"), 0, 1, 0, 1)
    spin1 = Gtk.SpinButton()
    spin1.set_adjustment(Gtk.Adjustment(0, 0, 100, 1))
    table.attach(spin1, 1, 2, 0, 1)
    table.attach(Gtk.Label(label="Gain:"), 0, 1, 1, 2)
    spin2 = Gtk.SpinButton()
    spin2.set_adjustment(Gtk.Adjustment(0, 0, 100, 1))
    table.attach(spin2, 1, 2, 1, 2)
    table.set_border_width(6)

    switchWithImage(image, table)

    def onValueChanged(spin):
        print(spin.get_value())

    spin1.connect("value-changed", onValueChanged)
    spin2.connect("value-changed", onValueChanged)

    win.show_all()
    win.connect("delete-event", Gtk.main_quit)
    Gtk.main()



"""Random Chess"""

import random

from pychess.Utils.const import RANDOMCHESS, VARIANTS_SHUFFLE
from pychess.Utils.Board import Board


class RandomBoard(Board):
    """:Description:
    * Randomly chosen pieces (two queens or three rooks possible)
    * Exactly one king of each color
    * Pieces placed randomly behind the pawns
    * No castling
    * Black's arrangement mirrors white's
    """

    variant = RANDOMCHESS
    __desc__ = _(
        "FICS wild/3: http://www.freechess.org/Help/HelpFiles/wild.html\n"
        + "* Randomly chosen pieces (two queens or three rooks possible)\n"
        + "* Exactly one king of each color\n"
        + "* Pieces placed randomly behind the pawns\n"
        + "* No castling\n"
        + "* Black's arrangement mirrors white's"
    )
    name = _("Random")
    cecp_name = "wild/3"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_SHUFFLE

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=self.random_start(), lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)

    def random_start(self):
        back_rank = random.sample(("r", "n", "b", "q") * 16, 7)
        back_rank.append("k")
        random.shuffle(back_rank)
        fen = "".join(back_rank)
        fen = fen + "/pppppppp/8/8/8/8/PPPPPPPP/" + fen.upper() + " w - - 0 1"
        return fen


if __name__ == "__main__":
    Board = RandomBoard(True)
    for i in range(10):
        print(Board.random_start())



import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from pychess import Variants
from pychess.Utils.const import (
    NORMALCHESS,
    ATOMICCHESS,
    BUGHOUSECHESS,
    CRAZYHOUSECHESS,
    LOSERSCHESS,
    SUICIDECHESS,
    FISCHERRANDOMCHESS,
    WILDCASTLESHUFFLECHESS,
    SHUFFLECHESS,
    RANDOMCHESS,
    ASYMMETRICRANDOMCHESS,
    WILDCASTLECHESS,
    UPSIDEDOWNCHESS,
    PAWNSPUSHEDCHESS,
    PAWNSPASSEDCHESS,
    GIVEAWAYCHESS,
    THREECHECKCHESS,
)

IC_CONNECTED, IC_DISCONNECTED = range(2)

# Fixed values of my relation to this game
# http://www.freechess.org/Help/HelpFiles/style12.html
(
    IC_POS_INITIAL,
    IC_POS_ISOLATED,
    IC_POS_OBSERVING_EXAMINATION,
    IC_POS_OP_TO_MOVE,
    IC_POS_OBSERVING,
    IC_POS_ME_TO_MOVE,
    IC_POS_EXAMINATING,
) = range(-4, 3)

# RatingType
(
    TYPE_BLITZ,
    TYPE_STANDARD,
    TYPE_LIGHTNING,
    TYPE_WILD,
    TYPE_BUGHOUSE,
    TYPE_CRAZYHOUSE,
    TYPE_SUICIDE,
    TYPE_LOSERS,
    TYPE_ATOMIC,
    TYPE_BULLET,
    TYPE_ONE_MINUTE,
    TYPE_THREE_MINUTE,
    TYPE_FIVE_MINUTE,
    TYPE_FIFTEEN_MINUTE,
    TYPE_FORTYFIVE_MINUTE,
    TYPE_CHESS960,
    TYPE_UNTIMED,
    TYPE_EXAMINED,
    TYPE_OTHER,
) = range(19)

RATING_TYPES = (
    TYPE_BLITZ,
    TYPE_STANDARD,
    TYPE_LIGHTNING,
    TYPE_BULLET,
    TYPE_ONE_MINUTE,
    TYPE_THREE_MINUTE,
    TYPE_FIVE_MINUTE,
    TYPE_FIFTEEN_MINUTE,
    TYPE_FORTYFIVE_MINUTE,
    TYPE_ATOMIC,
    TYPE_BUGHOUSE,
    TYPE_CRAZYHOUSE,
    TYPE_LOSERS,
    TYPE_SUICIDE,
    TYPE_WILD,
    TYPE_CHESS960,
    TYPE_UNTIMED,
)

# Rating deviations
DEVIATION_NONE, DEVIATION_ESTIMATED, DEVIATION_PROVISIONAL = range(3)

(
    IC_STATUS_PLAYING,
    IC_STATUS_ACTIVE,
    IC_STATUS_BUSY,
    IC_STATUS_OFFLINE,
    IC_STATUS_AVAILABLE,
    IC_STATUS_NOT_AVAILABLE,
    IC_STATUS_EXAMINING,
    IC_STATUS_IDLE,
    IC_STATUS_IN_TOURNAMENT,
    IC_STATUS_RUNNING_SIMUL_MATCH,
    IC_STATUS_UNKNOWN,
) = range(11)

TITLES_RE = r"(?:\([A-Z*]+\))*"
NAMES_RE = "[A-Za-z]+"

DEVIATION = {
    "E": DEVIATION_ESTIMATED,
    "P": DEVIATION_PROVISIONAL,
    " ": DEVIATION_NONE,
    "": DEVIATION_NONE,
}

STATUS = {
    "^": IC_STATUS_PLAYING,
    " ": IC_STATUS_AVAILABLE,
    ".": IC_STATUS_IDLE,
    "#": IC_STATUS_EXAMINING,
    ":": IC_STATUS_NOT_AVAILABLE,
    "~": IC_STATUS_RUNNING_SIMUL_MATCH,
    "&": IC_STATUS_IN_TOURNAMENT,
}


class GameType:
    def __init__(
        self,
        fics_name,
        short_fics_name,
        rating_type,
        display_text=None,
        variant_type=NORMALCHESS,
    ):
        self.fics_name = fics_name
        self.short_fics_name = short_fics_name
        self.rating_type = rating_type
        if display_text:
            self.display_text = display_text
        self.variant_type = variant_type

    @property
    def variant(self):
        return Variants.variants[self.variant_type]

    def __repr__(self):
        s = "<GameType "
        s += "fics_name='%s', " % self.fics_name
        s += "display_text='%s'>" % self.display_text
        return s


class NormalGameType(GameType):
    def __init__(self, fics_name, short_fics_name, rating_type, display_text):
        GameType.__init__(
            self, fics_name, short_fics_name, rating_type, display_text=display_text
        )


class VariantGameType(GameType):
    def __init__(self, fics_name, short_fics_name, rating_type, variant_type):
        GameType.__init__(
            self, fics_name, short_fics_name, rating_type, variant_type=variant_type
        )

    @property
    def display_text(self):
        assert self.variant_type is not None
        return Variants.variants[self.variant_type].name

    @property
    def seek_text(self):
        if "/" in self.fics_name:
            return self.fics_name.replace("/", " ")
        else:
            return self.fics_name


class WildGameType(VariantGameType):
    _instances = []

    def __init__(self, fics_name, variant_type):
        VariantGameType.__init__(
            self, fics_name, "w", TYPE_WILD, variant_type=variant_type
        )
        WildGameType._instances.append(self)

    @classmethod
    def instances(cls):
        return cls._instances


# FICS game types
GAME_TYPES = {
    "blitz": NormalGameType("blitz", "b", TYPE_BLITZ, _("Blitz")),
    "standard": NormalGameType("standard", "s", TYPE_STANDARD, _("Standard")),
    "lightning": NormalGameType("lightning", "l", TYPE_LIGHTNING, _("Lightning")),
    "1-minute": NormalGameType("1-minute", "o", TYPE_ONE_MINUTE, _("1-minute")),
    "3-minute": NormalGameType("3-minute", "M", TYPE_THREE_MINUTE, _("3-minute")),
    "5-minute": NormalGameType("5-minute", "f", TYPE_FIVE_MINUTE, _("5-minute")),
    "15-minute": NormalGameType("15-minute", "F", TYPE_FIFTEEN_MINUTE, _("15-minute")),
    "45-minute": NormalGameType(
        "45-minute", "J", TYPE_FORTYFIVE_MINUTE, _("45-minute")
    ),
    "chess960": NormalGameType("chess960", "K", TYPE_CHESS960, _("Chess960")),
    "untimed": NormalGameType("untimed", "u", TYPE_UNTIMED, _("Untimed")),
    "examined": NormalGameType("examined", "e", TYPE_EXAMINED, _("Examined")),
    "nonstandard": NormalGameType("nonstandard", "n", TYPE_OTHER, _("Other")),
    "w20": NormalGameType("loaded", "w20", TYPE_OTHER, _("Other")),  # loadfen/loadgame
    "w21": NormalGameType(
        "loaded", "w21", TYPE_OTHER, _("Other")
    ),  # thematic tournaments
    "atomic": VariantGameType("atomic", "x", TYPE_ATOMIC, ATOMICCHESS),
    "bughouse": VariantGameType("bughouse", "B", TYPE_BUGHOUSE, BUGHOUSECHESS),
    "crazyhouse": VariantGameType("crazyhouse", "z", TYPE_CRAZYHOUSE, CRAZYHOUSECHESS),
    "losers": VariantGameType("losers", "L", TYPE_LOSERS, LOSERSCHESS),
    "suicide": VariantGameType("suicide", "S", TYPE_SUICIDE, SUICIDECHESS),
    # FICS http://www.freechess.org/Help/HelpFiles/wild.html
    "wild/fr": WildGameType("wild/fr", FISCHERRANDOMCHESS),
    "wild/0": WildGameType("wild/0", WILDCASTLECHESS),
    "wild/1": WildGameType("wild/1", WILDCASTLESHUFFLECHESS),
    "wild/2": WildGameType("wild/2", SHUFFLECHESS),
    "wild/3": WildGameType("wild/3", RANDOMCHESS),
    "wild/4": WildGameType("wild/4", ASYMMETRICRANDOMCHESS),
    "wild/5": WildGameType("wild/5", UPSIDEDOWNCHESS),
    "wild/8": WildGameType("wild/8", PAWNSPUSHEDCHESS),
    "wild/8a": WildGameType("wild/8a", PAWNSPASSEDCHESS),
    # ICC https://www.chessclub.com/user/helpcenter/tips/wild.html
    "w17": WildGameType("w17", LOSERSCHESS),
    "w26": WildGameType("w26", GIVEAWAYCHESS),
    "w24": WildGameType("w24", BUGHOUSECHESS),
    "w23": WildGameType("w23", CRAZYHOUSECHESS),
    # "w16": WildGameType("w16", KRIEGSPIELCHESS),
    "w27": WildGameType("w27", ATOMICCHESS),
    # "w28": WildGameType("w28", SHATRANJCHESS),
    "w25": WildGameType("w25", THREECHECKCHESS),
    "w1": WildGameType("w1", WILDCASTLESHUFFLECHESS),
    "w2": WildGameType("w2", SHUFFLECHESS),
    "w3": WildGameType("w3", RANDOMCHESS),
    "w4": WildGameType("w4", ASYMMETRICRANDOMCHESS),
    "w22": WildGameType("w22", FISCHERRANDOMCHESS),
    "w5": WildGameType("w5", UPSIDEDOWNCHESS),
    # "w7": WildGameType("w7", THREEPAWNSCHESS),
    "w8": WildGameType("w8", PAWNSPUSHEDCHESS),
    # "w9": WildGameType("w9", TWOKINGSCHESS),
    # "w18": WildGameType("w18", EIGHTQUEENSCHESS),
    # "w19": WildGameType("w19", KNNKPCHESS),
}

VARIANT_GAME_TYPES = {}
for key in GAME_TYPES:
    if isinstance(GAME_TYPES[key], VariantGameType):
        VARIANT_GAME_TYPES[GAME_TYPES[key].variant_type] = GAME_TYPES[key]

# The following 3 GAME_TYPES_* data structures don't have any real entries
# for the WildGameType's in GAME_TYPES above, and instead use
# a dummy type for the all-encompassing "Wild" FICS rating for wild/* games
GAME_TYPES_BY_SHORT_FICS_NAME = {
    "w": GameType("wild", "w", TYPE_WILD, display_text=_("Wild"))
}
for key in GAME_TYPES:
    if not isinstance(GAME_TYPES[key], WildGameType):
        GAME_TYPES_BY_SHORT_FICS_NAME[GAME_TYPES[key].short_fics_name] = GAME_TYPES[key]

GAME_TYPES_BY_RATING_TYPE = {}
for key in GAME_TYPES_BY_SHORT_FICS_NAME:
    GAME_TYPES_BY_RATING_TYPE[GAME_TYPES_BY_SHORT_FICS_NAME[key].rating_type] = (
        GAME_TYPES_BY_SHORT_FICS_NAME[key]
    )

GAME_TYPES_BY_FICS_NAME = {}
for key in GAME_TYPES_BY_SHORT_FICS_NAME:
    GAME_TYPES_BY_FICS_NAME[GAME_TYPES_BY_SHORT_FICS_NAME[key].fics_name] = (
        GAME_TYPES_BY_SHORT_FICS_NAME[key]
    )


# Finally add conflicting ICC game type
GAME_TYPES["bullet"] = NormalGameType("bullet", "B", TYPE_BULLET, _("Bullet"))
GAME_TYPES_BY_FICS_NAME["bullet"] = GAME_TYPES["bullet"]
GAME_TYPES_BY_RATING_TYPE[TYPE_BULLET] = GAME_TYPES["bullet"]
# GAME_TYPES_BY_SHORT_FICS_NAME["B"] will be fixed in FICSConnections.py
# and VARIANT_GAME_TYPES[FISCHERRANDOMCHESS] also


def type_to_display_text(typename):
    if "loaded from" in typename.lower():
        typename = typename.split()[-1]
    if typename in GAME_TYPES:
        return GAME_TYPES[typename].display_text
    # Default solution for eco/A00 and a few others
    elif "/" in typename:
        a, b = typename.split("/")
        a = a[0].upper() + a[1:]
        b = b[0].upper() + b[1:]
        return a + " " + b
    else:
        # Otherwise forget about it
        return typename[0].upper() + typename[1:]


def time_control_to_gametype(minutes, gain):
    assert isinstance(minutes, int) and isinstance(gain, int)
    assert minutes >= 0 and gain >= 0
    gainminutes = gain > 0 and (gain * 60) - 1 or 0
    if minutes == 0 and gain == 0:
        return GAME_TYPES["untimed"]
    elif (minutes * 60) + gainminutes >= (15 * 60):
        return GAME_TYPES["standard"]
    elif (minutes * 60) + gainminutes >= (3 * 60):
        return GAME_TYPES["blitz"]
    else:
        return GAME_TYPES["lightning"]


(
    TYPE_ADMINISTRATOR,
    TYPE_BLINDFOLD,
    TYPE_COMPUTER,
    TYPE_TEAM,
    TYPE_UNREGISTERED,
    TYPE_CHESS_ADVISOR,
    TYPE_SERVICE_REPRESENTATIVE,
    TYPE_TOURNAMENT_DIRECTOR,
    TYPE_MAMER_MANAGER,
    TYPE_GRAND_MASTER,
    TYPE_INTERNATIONAL_MASTER,
    TYPE_FIDE_MASTER,
    TYPE_WOMAN_GRAND_MASTER,
    TYPE_WOMAN_INTERNATIONAL_MASTER,
    TYPE_WOMAN_FIDE_MASTER,
    TYPE_DUMMY_ACCOUNT,
    TYPE_CANDIDATE_MASTER,
    TYPE_FIDE_ARBEITER,
    TYPE_NATIONAL_MASTER,
    TYPE_DISPLAY_MASTER,
) = range(20)

TITLE_TYPE_DISPLAY_TEXTS = (
    _("Administrator"),
    _("Blindfold Account"),
    _("Computer"),
    _("Team Account"),
    _("Unregistered"),
    _("Chess Advisor"),
    _("Service Representative"),
    _("Tournament Director"),
    _("Mamer Manager"),
    _("Grand Master"),
    _("International Master"),
    _("FIDE Master"),
    _("Woman Grand Master"),
    _("Woman International Master"),
    _("Woman FIDE Master"),
    _("Dummy Account"),
    _("Candidate Master"),
    _("FIDE Arbeiter"),
    _("National Master"),
    _("Display Master"),
)

TITLE_TYPE_DISPLAY_TEXTS_SHORT = (
    _("*"),
    _("B"),
    _("C"),
    _("T"),
    _("U"),
    _("CA"),
    _("SR"),
    _("TD"),
    _("TM"),
    _("GM"),
    _("IM"),
    _("FM"),
    _("WGM"),
    _("WIM"),
    _("WFM"),
    _("D"),
    _("H"),
    _("CM"),
    _("FA"),
    _("NM"),
    _("DM"),
)

TITLES = {  # From FICS 'help who'
    "*": TYPE_ADMINISTRATOR,
    "B": TYPE_BLINDFOLD,
    "C": TYPE_COMPUTER,
    "T": TYPE_TEAM,
    "U": TYPE_UNREGISTERED,
    "CA": TYPE_CHESS_ADVISOR,
    "SR": TYPE_SERVICE_REPRESENTATIVE,
    "TD": TYPE_TOURNAMENT_DIRECTOR,
    "TM": TYPE_MAMER_MANAGER,
    "GM": TYPE_GRAND_MASTER,
    "IM": TYPE_INTERNATIONAL_MASTER,
    "FM": TYPE_FIDE_MASTER,
    "WFM": TYPE_WOMAN_FIDE_MASTER,
    "WIM": TYPE_WOMAN_INTERNATIONAL_MASTER,
    "WGM": TYPE_WOMAN_GRAND_MASTER,
    "D": TYPE_DUMMY_ACCOUNT,
    "H": TYPE_SERVICE_REPRESENTATIVE,
    "CM": TYPE_CANDIDATE_MASTER,
    "FA": TYPE_FIDE_ARBEITER,
    "NM": TYPE_NATIONAL_MASTER,
    "DM": TYPE_DISPLAY_MASTER,
}

HEX_TO_TITLE = {
    0x1: TYPE_UNREGISTERED,
    0x2: TYPE_COMPUTER,
    0x4: TYPE_GRAND_MASTER,
    0x8: TYPE_INTERNATIONAL_MASTER,
    0x10: TYPE_FIDE_MASTER,
    0x20: TYPE_WOMAN_GRAND_MASTER,
    0x40: TYPE_WOMAN_INTERNATIONAL_MASTER,
    0x80: TYPE_WOMAN_FIDE_MASTER,
}


def parse_title_hex(titlehex):
    titles = set()
    for key in HEX_TO_TITLE:
        if int(titlehex, 16) & key:
            titles.add(HEX_TO_TITLE[key])
    return titles


def parseRating(rating):
    if rating[0] == " ":
        rating = rating[1:]
    if rating[-1].isalpha():
        rating = rating[:-1]
    return int(rating) if rating.isdigit() else 0


def get_infobarmessage_content(player, text, gametype=None):
    content = Gtk.HBox()
    icon = Gtk.Image()
    icon.set_from_pixbuf(player.getIcon(size=32, gametype=gametype))
    content.pack_start(icon, False, False, 4)
    label = Gtk.Label()
    label.set_markup(player.getMarkup(gametype=gametype))
    content.pack_start(label, False, False, 0)
    label = Gtk.Label()
    label.set_markup(text)
    content.pack_start(label, False, False, 0)
    return content


def get_infobarmessage_content2(player, heading_text, message_text, gametype=None):
    hbox = Gtk.HBox()
    image = Gtk.Image()
    image.set_from_pixbuf(player.getIcon(size=24, gametype=gametype))
    hbox.pack_start(image, False, False, 0)
    label = Gtk.Label()
    markup = player.getMarkup(gametype=gametype, long_titles=False)
    label.set_markup(markup + heading_text)
    hbox.pack_start(label, False, False, 0)
    vbox = Gtk.VBox()
    vbox.pack_start(hbox, False, False, 0)
    label = Gtk.Label()
    label.props.xalign = 0
    label.props.xpad = 4
    label.props.justify = Gtk.Justification.LEFT
    label.props.wrap = True
    label.set_width_chars(70)
    label.set_text(message_text)
    vbox.pack_start(label, False, False, 5)
    return vbox


"""
Internal command codes used in FICS block mode
(see "help block_codes" and "help iv_block").
Used mostly by internal library functions.

BLOCK_ variables are message boundary markers.

BLKCMD_ variables are command codes.
"""

BLOCK_START = chr(21)  # \U
BLOCK_SEPARATOR = chr(22)  # \V
BLOCK_END = chr(23)  # \W
BLOCK_POSE_START = chr(24)  # \X
BLOCK_POSE_END = chr(25)  # \Y

BLKCMD_NULL = 0
BLKCMD_GAME_MOVE = 1
BLKCMD_ABORT = 10
BLKCMD_ACCEPT = 11
BLKCMD_ADDLIST = 12
BLKCMD_ADJOURN = 13
BLKCMD_ALLOBSERVERS = 14
BLKCMD_ASSESS = 15
BLKCMD_BACKWARD = 16
BLKCMD_BELL = 17
BLKCMD_BEST = 18
BLKCMD_BNAME = 19
BLKCMD_BOARDS = 20
BLKCMD_BSETUP = 21
BLKCMD_BUGWHO = 22
BLKCMD_CBEST = 23
BLKCMD_CLEARMESSAGES = 24
BLKCMD_CLRSQUARE = 25
BLKCMD_CONVERT_BCF = 26
BLKCMD_CONVERT_ELO = 27
BLKCMD_CONVERT_USCF = 28
BLKCMD_COPYGAME = 29
BLKCMD_CRANK = 30
BLKCMD_CSHOUT = 31
BLKCMD_DATE = 32
BLKCMD_DECLINE = 33
BLKCMD_DRAW = 34
BLKCMD_ECO = 35
BLKCMD_EXAMINE = 36
BLKCMD_FINGER = 37
BLKCMD_FLAG = 38
BLKCMD_FLIP = 39
BLKCMD_FMESSAGE = 40
BLKCMD_FOLLOW = 41
BLKCMD_FORWARD = 42
BLKCMD_GAMES = 43
BLKCMD_GETGI = 44
BLKCMD_GETPI = 45
BLKCMD_GINFO = 46
BLKCMD_GOBOARD = 47
BLKCMD_HANDLES = 48
BLKCMD_HBEST = 49
BLKCMD_HELP = 50
BLKCMD_HISTORY = 51
BLKCMD_HRANK = 52
BLKCMD_INCHANNEL = 53
BLKCMD_INDEX = 54
BLKCMD_INFO = 55
BLKCMD_ISET = 56
BLKCMD_IT = 57
BLKCMD_IVARIABLES = 58
BLKCMD_JKILL = 59
BLKCMD_JOURNAL = 60
BLKCMD_JSAVE = 61
BLKCMD_KIBITZ = 62
BLKCMD_LIMITS = 63
BLKCMD_LINE = 64  # Not on FICS
BLKCMD_LLOGONS = 65
BLKCMD_LOGONS = 66
BLKCMD_MAILHELP = 67
BLKCMD_MAILMESS = 68
BLKCMD_MAILMOVES = 69
BLKCMD_MAILOLDMOVES = 70
BLKCMD_MAILSOURCE = 71
BLKCMD_MAILSTORED = 72
BLKCMD_MATCH = 73
BLKCMD_MESSAGES = 74
BLKCMD_MEXAMINE = 75
BLKCMD_MORETIME = 76
BLKCMD_MOVES = 77
BLKCMD_NEWS = 78
BLKCMD_NEXT = 79
BLKCMD_OBSERVE = 80
BLKCMD_OLDMOVES = 81
BLKCMD_OLDSTORED = 82
BLKCMD_OPEN = 83
BLKCMD_PARTNER = 84
BLKCMD_PASSWORD = 85
BLKCMD_PAUSE = 86
BLKCMD_PENDING = 87
BLKCMD_PFOLLOW = 88
BLKCMD_POBSERVE = 89
BLKCMD_PREFRESH = 90
BLKCMD_PRIMARY = 91
BLKCMD_PROMOTE = 92
BLKCMD_PSTAT = 93
BLKCMD_PTELL = 94
BLKCMD_PTIME = 95
BLKCMD_QTELL = 96
BLKCMD_QUIT = 97
BLKCMD_RANK = 98
BLKCMD_RCOPYGAME = 99
BLKCMD_RFOLLOW = 100
BLKCMD_REFRESH = 101
BLKCMD_REMATCH = 102
BLKCMD_RESIGN = 103
BLKCMD_RESUME = 104
BLKCMD_REVERT = 105
BLKCMD_ROBSERVE = 106
BLKCMD_SAY = 107
BLKCMD_SERVERS = 108
BLKCMD_SET = 109
BLKCMD_SHOUT = 110
BLKCMD_SHOWLIST = 111
BLKCMD_SIMABORT = 112
BLKCMD_SIMALLABORT = 113
BLKCMD_SIMADJOURN = 114
BLKCMD_SIMALLADJOURN = 115
BLKCMD_SIMGAMES = 116
BLKCMD_SIMMATCH = 117
BLKCMD_SIMNEXT = 118
BLKCMD_SIMOBSERVE = 119
BLKCMD_SIMOPEN = 120
BLKCMD_SIMPASS = 121
BLKCMD_SIMPREV = 122
BLKCMD_SMOVES = 123
BLKCMD_SMPOSITION = 124
BLKCMD_SPOSITION = 125
BLKCMD_STATISTICS = 126
BLKCMD_STORED = 127
BLKCMD_STYLE = 128
BLKCMD_SWITCH = 130
BLKCMD_TAKEBACK = 131
BLKCMD_TELL = 132
BLKCMD_TIME = 133
BLKCMD_TOMOVE = 134
BLKCMD_TOURNSET = 135
BLKCMD_UNALIAS = 136
BLKCMD_UNEXAMINE = 137
BLKCMD_UNOBSERVE = 138
BLKCMD_UNPAUSE = 139
BLKCMD_UPTIME = 140
BLKCMD_USCF = 141
BLKCMD_USTAT = 142
BLKCMD_VARIABLES = 143
BLKCMD_WHENSHUT = 144
BLKCMD_WHISPER = 145
BLKCMD_WHO = 146
BLKCMD_WITHDRAW = 147
BLKCMD_WNAME = 148
BLKCMD_XKIBITZ = 149
BLKCMD_XTELL = 150
BLKCMD_XWHISPER = 151
BLKCMD_ZNOTIFY = 152
BLKCMD_REPLY = 153  # Not on FICS
BLKCMD_SUMMON = 154
BLKCMD_SEEK = 155
BLKCMD_UNSEEK = 156
BLKCMD_SOUGHT = 157
BLKCMD_PLAY = 158
BLKCMD_ALIAS = 159
BLKCMD_NEWBIES = 160
BLKCMD_SR = 161
BLKCMD_CA = 162
BLKCMD_TM = 163
BLKCMD_GETGAME = 164
BLKCMD_CCNEWSE = 165
BLKCMD_CCNEWSF = 166
BLKCMD_CCNEWSI = 167
BLKCMD_CCNEWSP = 168
BLKCMD_CCNEWST = 169
BLKCMD_CSNEWSE = 170
BLKCMD_CSNEWSF = 171
BLKCMD_CSNEWSI = 172
BLKCMD_CSNEWSP = 173
BLKCMD_CSNEWST = 174
BLKCMD_CTNEWSE = 175
BLKCMD_CTNEWSF = 176
BLKCMD_CTNEWSI = 177
BLKCMD_CTNEWSP = 178
BLKCMD_CTNEWST = 179
BLKCMD_CNEWS = 180
BLKCMD_SNEWS = 181
BLKCMD_TNEWS = 182
BLKCMD_RMATCH = 183
BLKCMD_RSTAT = 184
BLKCMD_CRSTAT = 185
BLKCMD_HRSTAT = 186
BLKCMD_GSTAT = 187

# Note admin codes start from 300.

BLKCMD_ERROR_BADCOMMAND = 512
BLKCMD_ERROR_BADPARAMS = 513
BLKCMD_ERROR_AMBIGUOUS = 514
BLKCMD_ERROR_RIGHTS = 515
BLKCMD_ERROR_OBSOLETE = 516
BLKCMD_ERROR_REMOVED = 517
BLKCMD_ERROR_NOTPLAYING = 518
BLKCMD_ERROR_NOSEQUENCE = 519
BLKCMD_ERROR_LENGTH = 520

LIMIT_BLKCMD_ERRORS = 500

FICS_COMMANDS = [
    "abort",
    "accept",
    "addlist",
    "adjourn",
    "alias",
    "allobservers",
    "assess",
    "backward",
    "bell",
    "best",
    "boards",
    "bsetup",
    "bugwho",
    "cbest",
    "clearmessages",
    "convert_bcf",
    "convert_elo",
    "convert_uscf",
    "copygame",
    "crank",
    "cshout",
    "date",
    "decline",
    "draw",
    "examine",
    "finger",
    "flag",
    "flip",
    "fmessage",
    "follow",
    "forward",
    "games",
    "gnotify",
    "goboard",
    "handles",
    "hbest",
    "help",
    "history",
    "hrank",
    "inchannel",
    "index",
    "info",
    "it",
    "jkill",
    "jsave",
    "kibitz",
    "limits",
    "llogons",
    "logons",
    "mailhelp",
    "mailmess",
    "mailmoves",
    "mailoldmoves",
    "mailsource",
    "mailstored",
    "match",
    "messages",
    "mexamine",
    "moretime",
    "moves",
    "news",
    "next",
    "observe",
    "oldmoves",
    "open",
    "password",
    "pause",
    "pending",
    "pfollow",
    "play",
    "pobserve",
    "promote",
    "pstat",
    "qtell",
    "quit",
    "rank",
    "refresh",
    "resign",
    "resume",
    "revert",
    "say",
    "seek",
    "servers",
    "set",
    "shout",
    "showlist",
    "simabort",
    "simallabort",
    "simadjourn",
    "simalladjourn",
    "simgames",
    "simmatch",
    "simnext",
    "simobserve",
    "simopen",
    "simpass",
    "simprev",
    "smoves",
    "smposition",
    "sought",
    "sposition",
    "statistics",
    "stored",
    "style",
    "sublist",
    "switch",
    "takeback",
    "tell",
    "time",
    "unalias",
    "unexamine",
    "unobserve",
    "unpause",
    "unseek",
    "uptime",
    "ustat",
    "variables",
    "whisper",
    "who",
    "withdraw",
    "xkibitz",
    "xtell",
    "xwhisper",
    "znotify",
]

FICS_HELP = [
    "_index",
    "abort",
    "abuse",
    "academy",
    "accept",
    "addlist",
    "addresses",
    "adjourn",
    "adjournments",
    "adjudicate",
    "adjudication",
    "adm_app",
    "adm_info",
    "adm_new",
    "admins",
    "alias",
    "allobservers",
    "assess",
    "atomic",
    "audiochat",
    "avail_vars",
    "backward",
    "bclock",
    "bell",
    "best",
    "blind",
    "blindfold",
    "blindh",
    "blitz",
    "block_codes",
    "bname",
    "boards",
    "brating",
    "bsetup",
    "bughouse",
    "bughouse_strat",
    "bugreport",
    "bugwho",
    "busy",
    "ca",
    "category",
    "cbest",
    "censor",
    "chan_1",
    "chan_4",
    "channel",
    "channel_list",
    "channels",
    "chess_adviser",
    "chess_advisor",
    "clearmessage",
    "clearmessages",
    "clock",
    "clocks",
    "clrsquare",
    "cls",
    "cls_info",
    "command",
    "commands",
    "commit",
    "computer_app",
    "computer_list",
    "computers",
    "confidentiality",
    "convert_bcf",
    "convert_elo",
    "convert_uscf",
    "copygame",
    "crank",
    "crazyhouse",
    "crazyhouse_strat",
    "credit",
    "crstat",
    "cshout",
    "csnewse",
    "csnewsf",
    "csnewsi",
    "csnewsp",
    "csnewst",
    "date",
    "decline",
    "disclaimer",
    "disconnection",
    "draw",
    "eco",
    "eggo",
    "email",
    "etime",
    "examine",
    "exl",
    "fen",
    "fics_faq",
    "fics_lingo",
    "finger",
    "flag",
    "flip",
    "fmessage",
    "follow",
    "formula",
    "forward",
    "fr",
    "fr_rules",
    "ftp_hints",
    "games",
    "games",
    "getgame",
    "getgi",
    "getpi",
    "ginfo",
    "glicko",
    "gnotify",
    "goboard",
    "handle",
    "handles",
    "hbest",
    "help",
    "highlight",
    "history",
    "hrank",
    "hrstat",
    "hstat",
    "icsdrone",
    "idlenotify",
    "inchannel",
    "index",
    "indexfile",
    "inetchesslib",
    "info",
    "intellegence",
    "interfaces",
    "intro_analysis",
    "intro_basics",
    "intro_general",
    "intro_information",
    "intro_moving",
    "intro_playing",
    "intro_settings",
    "intro_talking",
    "intro_welcome",
    "irc_help",
    "iset",
    "it",
    "iv_allresults",
    "iv_atomic",
    "iv_audiochat",
    "iv_block",
    "iv_boardinfo",
    "iv_compressmove",
    "iv_crazyhouse",
    "iv_defprompt",
    "iv_extascii",
    "iv_extuserinfo",
    "iv_fr",
    "iv_gameinfo",
    "iv_graph",
    "iv_list",
    "iv_lock",
    "iv_pendinfo",
    "iv_seekinfo",
    "iv_seekremove",
    "iv_startpos",
    "ivariables",
    "jkill",
    "journal",
    "jsave",
    "kibitz",
    "kiblevel",
    "lag",
    "lecture1",
    "lessons",
    "lightning",
    "limits",
    "links",
    "lists",
    "llogons",
    "logons",
    "losers",
    "losers_chess",
    "mailhelp",
    "mailmess",
    "mailmoves",
    "mailoldmoves",
    "mailstored",
    "mamer",
    "manual_usage",
    "manual_vars",
    "match",
    "meeting_1_followup",
    "meeting_1_long",
    "meeting_1_short",
    "meetings_index",
    "messages",
    "mexamine",
    "moretime",
    "motd",
    "motd_fri",
    "motd_help",
    "motd_mon",
    "motd_sat",
    "motd_sun",
    "motd_thu",
    "motd_tue",
    "motd_wed",
    "moves",
    "mule",
    "new_features",
    "newbie",
    "news",
    "next",
    "noescape",
    "noplay",
    "notes",
    "notify",
    "observe",
    "odds",
    "oldmoves",
    "oldpstat",
    "open",
    "partner",
    "password",
    "pause",
    "pending",
    "pfollow",
    "pgn",
    "ping",
    "play",
    "pobserve",
    "powericsfaq",
    "prefresh",
    "primary",
    "private",
    "promote",
    "pstat",
    "ptell",
    "ptime",
    "qtell",
    "quit",
    "rank",
    "rating_changes",
    "ratings",
    "rcopygame",
    "rd",
    "refresh",
    "register",
    "relay",
    "relay_operator",
    "rematch",
    "replay",
    "resign",
    "result",
    "resume",
    "revert",
    "rfollow",
    "rmatch",
    "robofics",
    "robserve",
    "rstat",
    "sabort",
    "say",
    "sdraw",
    "seek",
    "servers",
    "set",
    "setup",
    "shout",
    "shout_quota",
    "showadmins",
    "showlist",
    "showsrs",
    "simabort",
    "simadjourn",
    "simallabort",
    "simalladjourn",
    "simgames",
    "simmatch",
    "simnext",
    "simobserve",
    "simopen",
    "simpass",
    "simprev",
    "simuls",
    "skype",
    "smoves",
    "smposition",
    "sought",
    "spending",
    "sposition",
    "sr",
    "sr_info",
    "standard",
    "statistics",
    "stats",
    "stc",
    "stored",
    "style",
    "style12",
    "sublist",
    "suicide_chess",
    "summon",
    "switch",
    "system_alias",
    "takeback",
    "team",
    "teamgames",
    "tell",
    "time",
    "timeseal",
    "timeseal_mac",
    "timeseal_os2",
    "timeseal_unix",
    "timeseal_windows",
    "timezones",
    "tm",
    "tomove",
    "totals",
    "totals_info",
    "tournset",
    "town_meetings",
    "townmtg1",
    "unalias",
    "unexamine",
    "unobserve",
    "unpause",
    "unseek",
    "untimed",
    "uptime",
    "uscf",
    "uscf_faq",
    "ustat",
    "v_autoflag",
    "v_automail",
    "v_availinfo",
    "v_availmax",
    "v_availmin",
    "v_bell",
    "v_bugopen",
    "v_chanoff",
    "v_cshout",
    "v_ctell",
    "v_echo",
    "v_flip",
    "v_formula",
    "v_gin",
    "v_height",
    "v_highlight",
    "v_inc",
    "v_interface",
    "v_jprivate",
    "v_kibitz",
    "v_kiblevel",
    "v_language",
    "v_mailmess",
    "v_messreply",
    "v_notakeback",
    "v_notifiedby",
    "v_open",
    "v_pgn",
    "v_pin",
    "v_private",
    "v_prompt",
    "v_provshow",
    "v_ptime",
    "v_rated",
    "v_ropen",
    "v_seek",
    "v_shout",
    "v_silence",
    "v_simopen",
    "v_style",
    "v_tell",
    "v_time",
    "v_tolerance",
    "v_tourney",
    "v_tzone",
    "v_unobserve",
    "v_width",
    "variables",
    "wclock",
    "webpage",
    "whenshut",
    "whisper",
    "who",
    "wild",
    "withdraw",
    "wname",
    "wrating",
    "xkibitz",
    "xtell",
    "xwhisper",
    "zhouse",
    "znotify",
]



import os
import re
from ctypes import (
    byref,
    c_byte,
    c_char_p,
    c_int,
    c_uint,
    c_ulong,
    c_size_t,
    c_double,
    Structure,
    CDLL,
    CFUNCTYPE,
    POINTER,
)

from .bitboard import firstBit, clearBit
from .lmovegen import genAllMoves, genCheckEvasions
from pychess.Utils.const import WHITE, BLACK, DRAW, WHITEWON, BLACKWON
from pychess.System import conf
from pychess.System.prefix import getDataPrefix
from pychess.System.Log import log


class TbStats(Structure):
    _fields_ = [
        ("wdl_easy_hits", c_ulong * 2),
        ("wdl_hard_prob", c_ulong * 2),
        ("wdl_soft_prob", c_ulong * 2),
        ("wdl_cachesize", c_size_t),
        ("wdl_occupancy", c_double),
        ("dtm_easy_hits", c_ulong * 2),
        ("dtm_hard_prob", c_ulong * 2),
        ("dtm_soft_prob", c_ulong * 2),
        ("dtm_cachesize", c_size_t),
        ("dtm_occupancy", c_double),
        ("total_hits", c_ulong * 2),
        ("memory_hits", c_ulong * 2),
        ("drive_hits", c_ulong * 2),
        ("drive_miss", c_ulong * 2),
        ("bytes_read", c_ulong * 2),
        ("files_opened", c_ulong),
        ("memory_efficiency", c_double),
    ]


class EgtbGaviota:
    def __init__(self):
        self.libgtb = None
        self.initialized = False

        # Get a list of files in the tablebase folder.
        configuredTbPath = conf.get("egtb_path")
        tbPath = configuredTbPath or getDataPrefix()
        try:
            tbPathContents = os.listdir(tbPath)
        except OSError as e:
            if configuredTbPath:
                log.warning("Unable to open Gaviota TB folder: %s" % repr(e))
            return

        # Find files named *.gtb.cp# and pick the most common "#".
        # (This is the compression scheme; the library currently only uses one at a time.)
        schemeCount = [0] * 10
        for filename in tbPathContents:
            match = re.search(r"\.gtb\.cp(\d)$", filename)
            if match:
                schemeCount[int(match.group(1))] += 1
        compressionScheme = max(zip(schemeCount, range(10)))
        if compressionScheme[0] == 0:
            if configuredTbPath:
                log.warning(
                    "Could not find any Gaviota TB files in %s" % configuredTbPath
                )
            return
        compressionScheme = compressionScheme[1]

        # Locate and load the library.
        if not self._loadLibrary():
            return
        self._setupFunctionPrototypes()

        self.pathList = self.tbpaths_init()
        self.pathList = self.tbpaths_add(self.pathList, tbPath.encode())
        initInfo = self.tb_init(True, compressionScheme, self.pathList)
        self.initialized = self.tb_is_initialized() != 0
        if not self.initialized:
            log.warning(initInfo or "Failed to initialize Gaviota EGTB library")
            self.pathList = self.tbpaths_done(self.pathList)
            return
        elif initInfo:
            log.info(initInfo)

        # TODO: Set up a WDL cache area once the engine can use it.
        self.initialized &= self.tbcache_init(4 * 1024 * 1024, 0)
        if not self.initialized:
            log.warning("Failed to initialize Gaviota EGTB cache")
            self.tb_done()
            self.pathList = self.tbpaths_done(self.pathList)
            return

        self.availability = self.tb_availability()

    def _del(self):
        if self.initialized:
            self.tb_done()
            self.pathList = self.tbpaths_done(self.pathList)

    def supports(self, size):
        return self.initialized and (
            sum(size) <= 2 or (self.availability & (3 << (2 * sum(size) - 6))) != 0
        )

    def scoreAllMoves(self, board):
        result, depth = self.scoreGame(board, False, False)
        if result is None:
            return []

        scores = []
        gen = board.isChecked() and genCheckEvasions or genAllMoves
        for move in gen(board):
            board.applyMove(move)
            if not board.opIsChecked():
                result, depth = self.scoreGame(board, False, False)
                if result is None:
                    log.warning("An EGTB file does not have all its dependencies")
                    board.popMove()
                    return []
                scores.append((move, result, depth))
            board.popMove()

        def mateScore(mrd):
            if mrd[1] == DRAW:
                return 0
            absScore = 32767 - mrd[2]
            if (board.color == WHITE) ^ (mrd[1] == WHITEWON):
                return absScore
            return -absScore

        scores.sort(key=mateScore)
        return scores

    def scoreGame(self, board, omitDepth, probeSoft):
        stm = board.color
        epsq = board.enpassant or 64  # 64 is tb_NOSQUARE
        castles = (board.castling >> 2 & 3) | (board.castling << 2 & 12)
        tbinfo = c_uint()
        depth = c_uint()

        SqArray = c_uint * 65
        PcArray = c_byte * 65
        pc, sq = [], []
        for color in (WHITE, BLACK):
            sq.append(SqArray())
            pc.append(PcArray())
            i = 0
            bb = board.friends[color]
            while bb:
                b = firstBit(bb)
                bb = clearBit(bb, b)
                sq[-1][i] = b
                pc[-1][i] = board.arBoard[b]
                i += 1
            sq[-1][i] = 64  # tb_NOSQUARE, terminates the list
            pc[-1][i] = 0  # tb_NOPIECE,  terminates the list

        if omitDepth and probeSoft:
            ok = self.tb_probe_WDL_soft(
                stm,
                epsq,
                castles,
                sq[WHITE],
                sq[BLACK],
                pc[WHITE],
                pc[BLACK],
                byref(tbinfo),
            )
        elif omitDepth and not probeSoft:
            ok = self.tb_probe_WDL_hard(
                stm,
                epsq,
                castles,
                sq[WHITE],
                sq[BLACK],
                pc[WHITE],
                pc[BLACK],
                byref(tbinfo),
            )
        elif not omitDepth and probeSoft:
            ok = self.tb_probe_soft(
                stm,
                epsq,
                castles,
                sq[WHITE],
                sq[BLACK],
                pc[WHITE],
                pc[BLACK],
                byref(tbinfo),
                byref(depth),
            )
        elif not omitDepth and not probeSoft:
            ok = self.tb_probe_hard(
                stm,
                epsq,
                castles,
                sq[WHITE],
                sq[BLACK],
                pc[WHITE],
                pc[BLACK],
                byref(tbinfo),
                byref(depth),
            )

        resultMap = [DRAW, WHITEWON, BLACKWON]
        if not ok or not 0 <= tbinfo.value <= 2:
            return None, None
        result = resultMap[tbinfo.value]
        if omitDepth or result == DRAW:
            depth = None
        else:
            depth = depth.value
        return result, depth

    def _loadLibrary(self):
        libName = "libgaviotatb.so.1.0.1"
        try:
            self.libgtb = CDLL(libName)
        except OSError:
            log.warning("Failed to load Gaviota EGTB library %s" % libName)
            return None
        return self.libgtb

    # Prototypes from gtb-probe.h follow.

    def _setupFunctionPrototypes(self):
        def proto(name, returnType, *args):
            argTypes = map(lambda x: x[0], args)
            argNames = map(lambda x: x[1], args)
            funcType = CFUNCTYPE(returnType, *argTypes)
            paramFlags = tuple(zip([1] * len(args), argNames))
            setattr(self, name, funcType((name, self.libgtb), paramFlags))

        paths_t = POINTER(c_char_p)
        uip = POINTER(c_uint)
        ucp = POINTER(c_byte)

        proto(
            "tb_init",
            c_char_p,
            (c_int, "verbosity"),
            (c_int, "compression_scheme"),
            (paths_t, "paths"),
        )
        proto(
            "tb_restart",
            c_char_p,
            (c_int, "verbosity"),
            (c_int, "compression_scheme"),
            (paths_t, "paths"),
        )
        proto("tb_done", None)
        proto(
            "tb_probe_hard",
            c_int,
            (c_uint, "stm"),
            (c_uint, "epsq"),
            (c_uint, "castles"),
            (uip, "wSQ"),
            (uip, "bSQ"),
            (ucp, "wPC"),
            (ucp, "bPC"),
            (uip, "tbinfo"),
            (uip, "plies"),
        )
        proto(
            "tb_probe_soft",
            c_int,
            (c_uint, "stm"),
            (c_uint, "epsq"),
            (c_uint, "castles"),
            (uip, "wSQ"),
            (uip, "bSQ"),
            (ucp, "wPC"),
            (ucp, "bPC"),
            (uip, "tbinfo"),
            (uip, "plies"),
        )
        proto(
            "tb_probe_WDL_hard",
            c_int,
            (c_uint, "stm"),
            (c_uint, "epsq"),
            (c_uint, "castles"),
            (uip, "wSQ"),
            (uip, "bSQ"),
            (ucp, "wPC"),
            (ucp, "bPC"),
            (uip, "tbinfo"),
        )
        proto(
            "tb_probe_WDL_soft",
            c_int,
            (c_uint, "stm"),
            (c_uint, "epsq"),
            (c_uint, "castles"),
            (uip, "wSQ"),
            (uip, "bSQ"),
            (ucp, "wPC"),
            (ucp, "bPC"),
            (uip, "tbinfo"),
        )
        proto("tb_is_initialized", c_int)
        proto("tb_availability", c_uint)
        proto("tb_indexmemory", c_size_t)
        proto("tbcache_init", c_int, (c_size_t, "cache_mem"), (c_int, "wdl_fraction"))
        proto(
            "tbcache_restart", c_int, (c_size_t, "cache_mem"), (c_int, "wdl_fraction")
        )
        proto("tbcache_done", None)
        proto("tbcache_is_on", c_int)
        proto("tbcache_flush", None)
        proto("tbstats_reset", None)
        proto("tbstats_get", None, (POINTER(TbStats), "stats"))
        proto("tbpaths_init", paths_t)
        proto("tbpaths_add", paths_t, (paths_t, "ps"), (c_char_p, "newpath"))
        proto("tbpaths_done", paths_t, (paths_t, "ps"))
        proto("tbpaths_getmain", c_char_p)



from gi.repository import GObject

from pychess.ic import GAME_TYPES_BY_FICS_NAME
from pychess.ic.icc import DG_WHO_AM_I, CN_YFINGER
from pychess.ic.managers.FingerManager import FingerManager, FingerObject

ELO, DEVIATION, WINS, LOSSES, DRAWS, TOTAL, BESTELO, BESTTIME = range(8)

ICC_RATING_TYPE_MAP = {
    "Bul": "bullet",
    "Bli": "blitz",
    "Sta": "standard",
    "1-m": "1-minute",
    "3-m": "3-minute",
    "5-m": "5-minute",
    "15-": "15-minute",
    "45-": "45-minute",
    "Che": "chess960",
}


class ICCFingerManager(FingerManager):
    def __init__(self, connection):
        GObject.GObject.__init__(self)
        self.connection = connection

        self.connection.expect_dg_line(DG_WHO_AM_I, self.on_icc_who_am_i)
        self.connection.expect_cn_line(CN_YFINGER, self.on_icc_yfinger)

        self.connection.client.run_command("set-2 %s 1" % DG_WHO_AM_I)

    def on_icc_yfinger(self, data):
        finger = FingerObject()
        lines = data.split("\n")
        rating_lines = {}
        for line in lines:
            key, value = line.split(" ", 1)
            prefix = key[:3]
            # print(key, value, prefix, key[3:])
            if prefix in ICC_RATING_TYPE_MAP:
                if prefix not in rating_lines:
                    rating_lines[prefix] = [0] * 8
                if key[3:] == "Rat":
                    value = int(value)
                    rating_lines[prefix][ELO] = value
                elif key[3:] == "Win":
                    value = int(value)
                    rating_lines[prefix][WINS] = value
                elif key[3:] == "Draw":
                    value = int(value)
                    rating_lines[prefix][DRAWS] = value
                elif key[3:] == "Loss":
                    value = int(value)
                    rating_lines[prefix][LOSSES] = value
                elif key[3:] == "Need":
                    value = int(value)
                    rating_lines[prefix][DEVIATION] = value
                elif key[3:] == "Best":
                    bestelo, besttime = value.split(" ", 1)
                    rating_lines[prefix][BESTELO] = int(bestelo)
                    rating_lines[prefix][BESTTIME] = besttime
            elif key == "Name":
                finger.setName(value)

        for prefix in rating_lines:
            gametype = GAME_TYPES_BY_FICS_NAME[ICC_RATING_TYPE_MAP[prefix]]
            finger.setRating(gametype.rating_type, rating_lines[prefix])

        self.emit("fingeringFinished", finger)

    def on_icc_who_am_i(self, data):
        name, titles = data.split(" ", 1)
        self.connection.username = name

    def finger(self, user):
        self.connection.client.run_command("yfinger %s" % user)



import re
import signal
import sys
from threading import Thread

import pychess
from pychess.Players.PyChess import PyChess
from pychess.System import conf, fident
from pychess.Utils.book import getOpenings
from pychess.Utils.const import (
    NORMALCHESS,
    FEN_START,
    BLACK,
    FISCHERRANDOMCHESS,
    CRAZYHOUSECHESS,
    WILDCASTLESHUFFLECHESS,
    LOSERSCHESS,
    SUICIDECHESS,
    ATOMICCHESS,
    THREECHECKCHESS,
    KINGOFTHEHILLCHESS,
    ASEANCHESS,
    MAKRUKCHESS,
    CAMBODIANCHESS,
    SITTUYINCHESS,
    GIVEAWAYCHESS,
    HORDECHESS,
    RACINGKINGSCHESS,
    PLACEMENTCHESS,
    SCHESS,
    LIGHTBRIGADECHESS,
    WHITE,
)
from pychess.Utils.lutils.Benchmark import benchmark
from pychess.Utils.lutils.perft import perft
from pychess.Utils.lutils.LBoard import LBoard
from pychess.Utils.lutils.ldata import MAXPLY
from pychess.Utils.lutils import lsearch, leval
from pychess.Utils.lutils.lmove import parseSAN, parseAny, toSAN, ParsingError
from pychess.Utils.lutils.lmovegen import genAllMoves, genCaptures, genCheckEvasions
from pychess.Utils.lutils.validator import validateMove
from pychess.System.Log import log
from pychess.Variants.horde import HORDESTART
from pychess.Variants.placement import PLACEMENTSTART
from pychess.Variants.threecheck import THREECHECKSTART
from pychess.Variants.asean import (
    ASEANSTART,
    MAKRUKSTART,
    KAMBODIANSTART,
    SITTUYINSTART,
)
from pychess.Variants.seirawan import SCHESSSTART
from pychess.Variants.lightbrigade import LIGHTBRIGADESTART

if sys.platform != "win32":
    import readline

    readline.clear_history()

ASCII = sys.platform == "win32"


def get_input():
    return input()


class PyChessCECP(PyChess):
    def __init__(self):
        PyChess.__init__(self)
        self.board = LBoard(NORMALCHESS)
        self.board.applyFen(FEN_START)

        self.forced = False
        self.analyzing = False
        self.thread = None

        self.features = {
            "ping": 1,
            "setboard": 1,
            "playother": 1,
            "san": 1,
            "usermove": 1,
            "time": 1,
            "draw": 1,
            "sigint": 0,
            "sigterm": 0,
            "reuse": 1,
            "analyze": 1,
            "myname": "PyChess %s" % pychess.VERSION,
            "variants": "normal,wildcastle,nocastle,fischerandom,crazyhouse,light-brigade,"
            + "losers,suicide,giveaway,horde,atomic,racingkings,seirawan,"
            + "kingofthehill,3check,placement,asean,cambodian,makruk,sittuyin",
            "colors": 0,
            "ics": 0,
            "name": 0,
            "pause": 0,  # Unimplemented
            "nps": 0,  # Unimplemented
            "debug": 1,
            "memory": 0,  # Unimplemented
            "smp": 0,  # Unimplemented
            "egt": "gaviota",
            "option": "skipPruneChance -slider 0 0 100",
        }
        python = sys.executable.split("/")[-1]
        python_version = "%s.%s.%s" % sys.version_info[0:3]
        self.print(
            "# {} [{} {}]".format(self.features["myname"], python, python_version)
        )

    def handle_sigterm(self, *args):
        self.__stopSearching()
        sys.exit(0)

    def makeReady(self):
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, self.handle_sigterm)

    def run(self):
        while True:
            try:
                line = get_input()
            except EOFError:
                line = "quit"
            lines = line.split()

            try:
                if not lines:
                    continue

                log.debug(line, extra={"task": "xboard"})

                # CECP commands
                # See http://home.hccnet.nl/h.g.muller/engine-intf.html

                if lines[0] == "xboard":
                    pass

                elif lines[0] == "protover":
                    stringPairs = [
                        "=".join([k, '"%s"' % v if isinstance(v, str) else str(v)])
                        for k, v in self.features.items()
                    ]
                    self.print("feature %s" % " ".join(stringPairs))
                    self.print("feature done=1")

                elif lines[0] in ("accepted", "rejected"):
                    # We only really care about one case:
                    if tuple(lines) == ("rejected", "debug"):
                        self.debug = False

                elif lines[0] == "new":
                    self.__stopSearching()
                    self.board = LBoard(NORMALCHESS)
                    self.board.applyFen(FEN_START)
                    self.outOfBook = False
                    self.forced = False
                    self.playingAs = BLACK
                    self.clock[:] = self.basetime, self.basetime
                    self.searchtime = 0
                    self.sd = MAXPLY
                    if self.analyzing:
                        self.__analyze()

                elif lines[0] == "variant":
                    if len(lines) > 1:
                        if lines[1] == "fischerandom":
                            self.board.variant = FISCHERRANDOMCHESS
                        elif lines[1] == "crazyhouse":
                            self.board.variant = CRAZYHOUSECHESS
                            self.board.iniHouse()
                        elif lines[1] == "wildcastle":
                            self.board.variant = WILDCASTLESHUFFLECHESS
                        elif lines[1] == "losers":
                            self.board.variant = LOSERSCHESS
                        elif lines[1] == "suicide":
                            self.board.variant = SUICIDECHESS
                        elif lines[1] == "giveaway":
                            self.board.variant = GIVEAWAYCHESS
                        elif lines[1] == "atomic":
                            self.board.variant = ATOMICCHESS
                            self.board.iniAtomic()
                        elif lines[1] == "3check":
                            self.board = LBoard(THREECHECKCHESS)
                            self.board.applyFen(THREECHECKSTART)
                        elif lines[1] == "racingkings":
                            self.board.variant = RACINGKINGSCHESS
                        elif lines[1] == "kingofthehill":
                            self.board.variant = KINGOFTHEHILLCHESS
                        elif lines[1] == "horde":
                            self.board = LBoard(HORDECHESS)
                            self.board.applyFen(HORDESTART)
                        elif lines[1] == "placement":
                            self.board = LBoard(PLACEMENTCHESS)
                            self.board.applyFen(PLACEMENTSTART)
                        elif lines[1] == "asean":
                            self.board = LBoard(ASEANCHESS)
                            self.board.applyFen(ASEANSTART)
                        elif lines[1] == "makruk":
                            self.board = LBoard(MAKRUKCHESS)
                            self.board.applyFen(MAKRUKSTART)
                        elif lines[1] == "cambodian":
                            self.board = LBoard(CAMBODIANCHESS)
                            self.board.applyFen(KAMBODIANSTART)
                        elif lines[1] == "sittuyin":
                            self.board = LBoard(SITTUYINCHESS)
                            self.board.applyFen(SITTUYINSTART)
                        elif lines[1] == "seirawan":
                            self.board = LBoard(SCHESS)
                            self.board.applyFen(SCHESSSTART)
                        elif lines[1] == "light-brigade":
                            self.board = LBoard(LIGHTBRIGADECHESS)
                            self.board.applyFen(LIGHTBRIGADESTART)

                elif lines[0] == "quit":
                    self.forced = True
                    self.__stopSearching()
                    sys.exit(0)

                elif lines[0] == "random":
                    leval.random = True

                elif lines[0] == "force":
                    if not self.forced and not self.analyzing:
                        self.forced = True
                        self.__stopSearching()

                elif lines[0] == "go":
                    self.playingAs = self.board.color
                    self.forced = False
                    self.__go()

                elif lines[0] == "playother":
                    self.playingAs = 1 - self.board.color
                    self.forced = False
                    # TODO: start pondering, if possible

                elif lines[0] in ("black", "white"):
                    newColor = lines[0] == "black" and BLACK or WHITE
                    self.__stopSearching()
                    self.playingAs = 1 - newColor
                    if self.board.color != newColor:
                        self.board.setColor(newColor)
                        self.board.setEnpassant(None)
                    if self.analyzing:
                        self.__analyze()

                elif lines[0] == "level":
                    self.movestogo = int(lines[1])
                    inc = int(lines[3])
                    minutes = lines[2].split(":")
                    # Per protocol spec, strip off any non-numeric suffixes.
                    for i in range(len(minutes)):
                        minutes[i] = re.match(r"\d*", minutes[i]).group()
                    self.basetime = int(minutes[0]) * 60
                    if len(minutes) > 1 and minutes[1]:
                        self.basetime += int(minutes[1])
                    self.clock[:] = self.basetime, self.basetime
                    self.increment = inc
                    self.searchtime = 0

                elif lines[0] == "st":
                    self.searchtime = float(lines[1])

                elif lines[0] == "sd":
                    self.sd = int(lines[1])

                # Unimplemented: nps

                elif lines[0] == "time":
                    self.clock[self.playingAs] = float(lines[1]) / 100.0

                elif lines[0] == "otim":
                    self.clock[1 - self.playingAs] = float(lines[1]) / 100.0

                elif lines[0] == "usermove":
                    self.__stopSearching()
                    try:
                        move = parseAny(self.board, lines[1])
                    except ParsingError:
                        self.print("Error (unknown command): %s" % lines[1])
                        self.print(self.board.prepr(ascii=ASCII))
                        continue
                    if not validateMove(self.board, move):
                        self.print("Illegal move: %s" % lines[1])
                        self.print(self.board.prepr(ascii=ASCII))
                        continue
                    self.board.applyMove(move)
                    self.playingAs = self.board.color
                    if not self.forced and not self.analyzing:
                        self.__go()
                    if self.analyzing:
                        self.__analyze()

                elif lines[0] == "?":
                    if not self.forced and not self.analyzing:
                        self.__stopSearching()

                elif lines[0] == "ping":
                    self.print("pong %s" % lines[1])

                elif lines[0] == "draw":
                    if self.__willingToDraw():
                        self.print("offer draw")

                elif lines[0] == "result":
                    # We don't really care what the result is at the moment.
                    pass

                elif lines[0] == "setboard":
                    self.__stopSearching()
                    try:
                        self.board = LBoard(self.board.variant)
                        fen = " ".join(lines[1:])
                        self.board.applyFen(fen.replace("[", "/").replace("]", ""))
                    except SyntaxError as err:
                        self.print("tellusererror Illegal position: %s" % str(err))

                # "edit" is unimplemented. See docs. Exiting edit mode returns to analyze mode.

                elif lines[0] == "hint":
                    pass  # TODO: Respond "Hint: MOVE" if we have an expected reply

                elif lines[0] == "bk":
                    entries = getOpenings(self.board)
                    if entries:
                        totalWeight = sum(entry[1] for entry in entries)
                        for entry in entries:
                            self.print(
                                "\t%s\t%02.2f%%"
                                % (
                                    toSAN(self.board, entry[0]),
                                    entry[1] * 100.0 / totalWeight,
                                )
                            )

                elif lines[0] == "undo":
                    self.__stopSearching()
                    self.board.popMove()
                    if self.analyzing:
                        self.__analyze()

                elif lines[0] == "remove":
                    self.__stopSearching()
                    self.board.popMove()
                    self.board.popMove()
                    if self.analyzing:
                        self.__analyze()

                elif lines[0] in ("hard", "easy"):
                    self.ponder = lines[0] == "hard"

                elif lines[0] in ("post", "nopost"):
                    self.post = lines[0] == "post"

                elif lines[0] == "analyze":
                    self.analyzing = True
                    self.__analyze()

                elif lines[0] in ("name", "rating", "ics", "computer"):
                    pass  # We don't care.

                # Unimplemented: pause, resume

                elif lines[0] == "memory":
                    # FIXME: this is supposed to control the *total* memory use.
                    if lsearch.searching:
                        self.print("Error (already searching):", line)
                    else:
                        limit = int(lines[1])
                        if limit < 1:
                            self.print("Error (limit too low):", line)
                        else:
                            pass
                            # TODO implement
                            # lsearch.setHashSize(limit)

                elif lines[0] == "cores":
                    pass  # We aren't SMP-capable.

                elif lines[0] == "egtpath":
                    if len(lines) >= 3 and lines[1] == "gaviota":
                        if lines[2]:
                            conf.set("egtb_path", lines[2])
                        else:
                            conf.set("egtb_path", conf.get("egtb_path"))
                        from pychess.Utils.lutils.lsearch import enableEGTB

                        enableEGTB()

                elif lines[0] == "option" and len(lines) > 1:
                    name, eq, value = lines[1].partition("=")
                    if value:
                        value = int(
                            value
                        )  # CECP spec says option values are *always* numeric
                    if name == "skipPruneChance":
                        if 0 <= value <= 100:
                            self.skipPruneChance = value / 100.0
                        else:
                            self.print(
                                "Error (argument must be an integer 0..100): %s" % line
                            )

                # CECP analyze mode commands
                # See http://www.gnu.org/software/xboard/engine-intf.html#11

                elif lines[0] == "exit":
                    if self.analyzing:
                        self.__stopSearching()
                        self.analyzing = False

                # Periodic updates (".") are not implemented.

                # Custom commands

                elif lines[0] == "moves":
                    self.print(self.board.prepr(ascii=ASCII))
                    self.print(
                        [toSAN(self.board, move) for move in genAllMoves(self.board)]
                    )

                elif lines[0] == "captures":
                    self.print(self.board.prepr(ascii=ASCII))
                    self.print(
                        [toSAN(self.board, move) for move in genCaptures(self.board)]
                    )

                elif lines[0] == "evasions":
                    self.print(self.board.prepr(ascii=ASCII))
                    self.print(
                        [
                            toSAN(self.board, move)
                            for move in genCheckEvasions(self.board)
                        ]
                    )

                elif lines[0] == "benchmark":
                    if len(lines) > 1:
                        benchmark(int(lines[1]))
                    else:
                        benchmark()

                elif lines[0] == "profile":
                    if len(lines) > 1:
                        import cProfile

                        cProfile.runctx("benchmark()", locals(), globals(), lines[1])
                    else:
                        self.print("Usage: profile outputfilename")

                elif lines[0] == "perft":
                    root = "0" if len(lines) < 3 else lines[2]
                    depth = "1" if len(lines) == 1 else lines[1]
                    if root.isdigit() and depth.isdigit():
                        perft(self.board, int(depth), int(root))
                    else:
                        self.print("Error (arguments must be integer")

                elif lines[0] == "stop_unittest":
                    break

                elif len(lines) == 1:
                    # A GUI without usermove support might try to send a move.
                    try:
                        move = parseAny(self.board, line)
                    except ParsingError:
                        self.print("Error (unknown command): %s" % line)
                        continue
                    if not validateMove(self.board, move):
                        self.print("Illegal move: %s" % lines[0])
                        self.print(self.board.prepr(ascii=ASCII))
                        continue
                    self.__stopSearching()
                    self.board.applyMove(move)
                    self.playingAs = self.board.color
                    if not self.forced and not self.analyzing:
                        self.__go()
                    if self.analyzing:
                        self.__analyze()

                else:
                    self.print("Error (unknown command): %s" % line)
            except IndexError:
                self.print("Error (missing argument): %s" % line)

    def __stopSearching(self):
        lsearch.searching = False
        if self.thread:
            self.thread.join()

    def __go(self):
        def ondone(result):
            if not self.forced:
                self.board.applyMove(parseSAN(self.board, result))
                self.print("move %s" % result)
            # TODO: start pondering, if enabled

        self.thread = Thread(
            target=PyChess._PyChess__go,
            name=fident(PyChess._PyChess__go),
            args=(self, ondone),
        )
        self.thread.daemon = True
        self.thread.start()

    def __analyze(self):
        self.thread = Thread(
            target=PyChess._PyChess__analyze,
            name=fident(PyChess._PyChess__analyze),
            args=(self,),
        )
        self.thread.daemon = True
        self.thread.start()

    def __willingToDraw(self):
        return self.scr <= 0  # FIXME: this misbehaves in all but the simplest use cases



from gi.repository import Gtk, GObject

from pychess.Utils.const import ASEAN_VARIANTS, NORMALCHESS
from pychess.gfx import Pieces


class PieceWidget(Gtk.DrawingArea):
    def __init__(self, piece, variant=NORMALCHESS):
        GObject.GObject.__init__(self)
        self.connect("draw", self.expose)
        self.piece = piece
        self.variant = variant
        self.asean = variant in ASEAN_VARIANTS

    def setPiece(self, piece):
        self.piece = piece

    def getPiece(self):
        return self.piece

    def expose(self, widget, ctx):
        context = widget.get_window().cairo_create()
        rect = self.get_allocation()
        s_min = min(rect.width, rect.height)
        x_loc = (rect.width - s_min) / 2.0
        y_loc = (rect.height - s_min) / 2.0
        Pieces.drawPiece(
            self.piece,
            context,
            x_loc,
            y_loc,
            s_min,
            asean=self.asean,
            variant=self.variant,
        )



# S-Chess

from pychess.Utils.const import (
    SCHESS,
    VARIANTS_OTHER_NONSTANDARD,
    KNIGHT_PROMOTION,
    BISHOP_PROMOTION,
    ROOK_PROMOTION,
    QUEEN_PROMOTION,
    HAWK_PROMOTION,
    ELEPHANT_PROMOTION,
)
from pychess.Utils.Board import Board

SCHESSSTART = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR[heHE] w KQBCDFGkqbcdfg - 0 1"


class SchessBoard(Board):
    variant = SCHESS
    __desc__ = _("S-Chess: https://en.wikipedia.org/wiki/Seirawan_chess")
    name = _("S-Chess")
    cecp_name = "seirawan"
    need_initial_board = True
    standard_rules = False
    variant_group = VARIANTS_OTHER_NONSTANDARD
    PROMOTIONS = (
        KNIGHT_PROMOTION,
        BISHOP_PROMOTION,
        ROOK_PROMOTION,
        QUEEN_PROMOTION,
        HAWK_PROMOTION,
        ELEPHANT_PROMOTION,
    )

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=SCHESSSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



from pychess.Utils.const import WHITE, WHITEWON, BLACK, BLACKWON, DRAW


def get_elo_rating_change(model, overridden_welo, overridden_belo):
    """http://www.fide.com/fide/handbook.html?id=197&view=article (§8.5, July 2017)"""

    def individual_elo_change(elo_player, elo_opponent, blitz):
        result = {}

        # Adaptation of the inbound parameters
        pprov = "?" in elo_player
        try:
            pval = int(elo_player.replace("?", ""))
        except ValueError:
            pval = 0
        try:
            oval = int(elo_opponent.replace("?", ""))
        except ValueError:
            oval = 0
        if pval == 0 or oval == 0:
            return None

        # Development coefficient - We ignore the age of the player and we assume that
        # he is already rated. The provisional flag '?' just denotes that he has not
        # played his first 30 games, but it may also denotes that he never had any
        # ranking. The calculation being based on the current game only, we can't
        # handle that second specific case
        if blitz:
            k = 20
        else:
            if pprov:
                k = 40
            else:
                if pval >= 2400:
                    k = 10
                else:
                    k = 20

        # Probability of gain
        d = pval - oval
        d = max(-400, d)
        d = min(400, d)
        # The approximate formula should not be used : result["pd"] = 1.0/(1+10**(-d/400))
        pd = [
            50,
            50,
            50,
            50,
            51,
            51,
            51,
            51,
            51,
            51,
            51,
            52,
            52,
            52,
            52,
            52,
            52,
            52,
            53,
            53,
            53,
            53,
            53,
            53,
            53,
            53,
            54,
            54,
            54,
            54,
            54,
            54,
            54,
            55,
            55,
            55,
            55,
            55,
            55,
            55,
            56,
            56,
            56,
            56,
            56,
            56,
            56,
            57,
            57,
            57,
            57,
            57,
            57,
            57,
            58,
            58,
            58,
            58,
            58,
            58,
            58,
            58,
            59,
            59,
            59,
            59,
            59,
            59,
            59,
            60,
            60,
            60,
            60,
            60,
            60,
            60,
            60,
            61,
            61,
            61,
            61,
            61,
            61,
            61,
            62,
            62,
            62,
            62,
            62,
            62,
            62,
            62,
            63,
            63,
            63,
            63,
            63,
            63,
            63,
            64,
            64,
            64,
            64,
            64,
            64,
            64,
            64,
            65,
            65,
            65,
            65,
            65,
            65,
            65,
            66,
            66,
            66,
            66,
            66,
            66,
            66,
            66,
            67,
            67,
            67,
            67,
            67,
            67,
            67,
            67,
            68,
            68,
            68,
            68,
            68,
            68,
            68,
            68,
            69,
            69,
            69,
            69,
            69,
            69,
            69,
            69,
            70,
            70,
            70,
            70,
            70,
            70,
            70,
            70,
            71,
            71,
            71,
            71,
            71,
            71,
            71,
            71,
            71,
            72,
            72,
            72,
            72,
            72,
            72,
            72,
            72,
            73,
            73,
            73,
            73,
            73,
            73,
            73,
            73,
            73,
            74,
            74,
            74,
            74,
            74,
            74,
            74,
            74,
            74,
            75,
            75,
            75,
            75,
            75,
            75,
            75,
            75,
            75,
            76,
            76,
            76,
            76,
            76,
            76,
            76,
            76,
            76,
            77,
            77,
            77,
            77,
            77,
            77,
            77,
            77,
            77,
            78,
            78,
            78,
            78,
            78,
            78,
            78,
            78,
            78,
            78,
            79,
            79,
            79,
            79,
            79,
            79,
            79,
            79,
            79,
            79,
            80,
            80,
            80,
            80,
            80,
            80,
            80,
            80,
            80,
            80,
            81,
            81,
            81,
            81,
            81,
            81,
            81,
            81,
            81,
            81,
            81,
            82,
            82,
            82,
            82,
            82,
            82,
            82,
            82,
            82,
            82,
            82,
            83,
            83,
            83,
            83,
            83,
            83,
            83,
            83,
            83,
            83,
            83,
            84,
            84,
            84,
            84,
            84,
            84,
            84,
            84,
            84,
            84,
            84,
            84,
            85,
            85,
            85,
            85,
            85,
            85,
            85,
            85,
            85,
            85,
            85,
            85,
            86,
            86,
            86,
            86,
            86,
            86,
            86,
            86,
            86,
            86,
            86,
            86,
            86,
            87,
            87,
            87,
            87,
            87,
            87,
            87,
            87,
            87,
            87,
            87,
            87,
            87,
            88,
            88,
            88,
            88,
            88,
            88,
            88,
            88,
            88,
            88,
            88,
            88,
            88,
            88,
            88,
            88,
            89,
            89,
            89,
            89,
            89,
            89,
            89,
            89,
            89,
            89,
            89,
            89,
            89,
            90,
            90,
            90,
            90,
            90,
            90,
            90,
            90,
            90,
            90,
            90,
            90,
            90,
            90,
            90,
            90,
            90,
            91,
            91,
            91,
            91,
            91,
            91,
            91,
            91,
            91,
            91,
            91,
            91,
            91,
            91,
            91,
            91,
            91,
            92,
            92,
            92,
            92,
            92,
            92,
            92,
            92,
            92,
        ][abs(d)] / 100.0
        result["pd"] = pd if pval >= oval else 1.0 - pd

        # New difference in Elo for loss, draw, win
        for score in [0, 1, 2]:
            result["diff%d" % score] = round(k * ([0, 0.5, 1][score] - result["pd"]), 1)

        # Result
        return result

    # Gathering of the data
    welo = model.tags["WhiteElo"] if overridden_welo is None else overridden_welo
    belo = model.tags["BlackElo"] if overridden_belo is None else overridden_belo
    blitz = model.timemodel.isBlitzFide()

    # Result
    result = [None, None]
    result[WHITE] = individual_elo_change(welo, belo, blitz)
    result[BLACK] = individual_elo_change(belo, welo, blitz)
    return None if result[WHITE] is None or result[BLACK] is None else result


def get_elo_rating_change_str(model, player, overridden_welo, overridden_belo):
    """Determination of the ELO rating change"""

    erc = get_elo_rating_change(model, overridden_welo, overridden_belo)
    if erc is None:
        return ""
    erc = erc[player]

    # Status of the game
    if (model.status == WHITEWON and player == WHITE) or (
        model.status == BLACKWON and player == BLACK
    ):
        d = 2
    else:
        if (model.status == WHITEWON and player == BLACK) or (
            model.status == BLACKWON and player == WHITE
        ):
            d = 0
        else:
            if model.status == DRAW:
                d = 1
            else:
                return "{:.0f}%, {:.1f} / {:.1f} / {:.1f}".format(
                    100 * erc["pd"], erc["diff0"], erc["diff1"], erc["diff2"]
                )

    # Result
    return "{}{:.1f}".format("+" if erc["diff%d" % d] > 0 else "", erc["diff%d" % d])


def get_elo_rating_change_pgn(model, player):
    # One move must occur to validate the rating
    if model.ply == 0:
        return ""

    # Retrieval of the statistics for the player
    data = get_elo_rating_change(model, None, None)
    if data is None:
        return ""
    data = data[player]

    # Status of the game
    if (model.status == WHITEWON and player == WHITE) or (
        model.status == BLACKWON and player == BLACK
    ):
        d = 2
    else:
        if (model.status == WHITEWON and player == BLACK) or (
            model.status == BLACKWON and player == WHITE
        ):
            d = 0
        else:
            if model.status == DRAW:
                d = 1
            else:
                return ""

    # Result is rounded to the nearest integer
    r = int(round(data["diff%s" % d], 0))
    return "+%d" % r if r > 0 else str(r)



"""Queen Odds Variant"""

from pychess.Utils.const import QUEENODDSCHESS, VARIANTS_ODDS
from pychess.Utils.Board import Board

QUEENODDSSTART = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNB1KBNR w KQkq - 0 1"


class QueenOddsBoard(Board):
    """:Description: Standard chess rules but one side starts without a queen"""

    variant = QUEENODDSCHESS
    __desc__ = _("One player starts with one less queen piece")
    name = _("Queen odds")
    cecp_name = "normal"
    need_initial_board = True
    standard_rules = True
    variant_group = VARIANTS_ODDS

    def __init__(self, setup=False, lboard=None):
        if setup is True:
            Board.__init__(self, setup=QUEENODDSSTART, lboard=lboard)
        else:
            Board.__init__(self, setup=setup, lboard=lboard)



from gi.repository import GObject, Gtk


def get_message_content(heading_text, message_text, image_stock_id):
    label = Gtk.Label()
    label.props.xalign = 0
    label.props.justify = Gtk.Justification.LEFT
    label.props.wrap = True
    label.set_text(f"{heading_text} {message_text}")
    return label


class InfoBarMessageButton(GObject.GObject):
    def __init__(self, text, response_id, sensitive=True, tooltip_text=""):
        GObject.GObject.__init__(self)
        self.text = text
        self.response_id = response_id
        self.sensitive = sensitive
        self.tooltip_text = tooltip_text
        self._button = None

    def get_sensitive(self):
        return self._sensitive

    def set_sensitive(self, sensitive):
        self._sensitive = sensitive

    sensitive = GObject.property(get_sensitive, set_sensitive)

    def get_tooltip_text(self):
        return self._tooltip_text

    def set_tooltip_text(self, tooltip_text):
        self._tooltip_text = tooltip_text

    tooltip_text = GObject.property(get_tooltip_text, set_tooltip_text)


class InfoBarMessage(Gtk.InfoBar):
    __gsignals__ = {
        "dismissed": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, message_type, content, callback):
        GObject.GObject.__init__(self)
        self.callback = callback
        self.content = content
        self.buttons = []

        self.get_content_area().add(content)

    def add_button(self, button):
        """
        All buttons must be added before doing InfoBarNotebook.push_message()
        """
        if not isinstance(button, InfoBarMessageButton):
            raise TypeError("Not an InfoBarMessageButton: %s" % repr(button))
        self.buttons.append(button)
        button._button = Gtk.InfoBar.add_button(self, button.text, button.response_id)

        def set_sensitive(button, property):
            if self.get_children():
                self.set_response_sensitive(button.response_id, button.sensitive)

        button.connect("notify::sensitive", set_sensitive)

        def set_tooltip_text(button, property):
            button._button.set_property("tooltip-text", button.tooltip_text)

        button.connect("notify::tooltip-text", set_tooltip_text)

    def dismiss(self):
        self.hide()
        self.emit("dismissed")

    def update_content(self, content):
        for widget in self.get_content_area():
            self.get_content_area().remove(widget)
        self.get_content_area().add(content)
        self.show_all()


class InfoBarNotebook(Gtk.Notebook):
    """
    This is a :class:`Gtk.Notebook` which manages InfoBarMessage objects pushed onto it via
    push_message() like a stack. If/when the current message at the top of the
    stack is responded to or dismissed by the user, the next message in the
    stack waiting for a response is displayed. Messages that aren't applicable
    anymore can be removed from anywhere in the InfoBar message stack by calling
    message.dismiss()
    """

    def __init__(self, name=None):
        Gtk.Notebook.__init__(self)
        if name is not None:
            self.set_name(name)
        self.get_tab_label_text = self.customGetTabLabelText
        self.set_show_tabs(False)

    def customGetTabLabelText(self, child):
        return child.content.get_text()

    def push_message(self, message):
        def on_dismissed(mesage):
            page_num = self.page_num(message)
            if page_num != -1:
                self.remove_page(page_num)

        def onResponse(message, response_id):
            if callable(message.callback):
                message.callback(self, response_id, message)
            page_num = self.page_num(message)
            if page_num != -1:
                self.remove_page(page_num)

        current_page = self.get_current_page()
        if current_page > 0:
            self.remove_page(current_page)
        self.append_page(message, None)
        message.connect("response", onResponse)
        message.connect("dismissed", on_dismissed)
        self.show_all()

    def clear_messages(self):
        for child in self.get_children():
            self.remove(child)
        self.show_all()



import asyncio
import os
import os.path
import gettext
import locale
import sys

from math import pi
from operator import attrgetter
from itertools import groupby
from io import StringIO

import gi

try:
    gi.require_version("Gdk", "3.0")
    gi.require_version("GLib", "2.0")
    gi.require_version("GObject", "2.0")
    gi.require_version("Gtk", "3.0")
    from gi.repository import Gdk, Gtk, GLib, GObject

    gi.require_version("GtkSource", "4")
    from gi.repository import GtkSource

    GtkSource.init()
except Exception:
    print("Failed to import required gi module version")
    sys.exit(1)

from pychess.Utils.IconLoader import load_icon, get_pixbuf
from pychess.Utils.GameModel import GameModel
from pychess.Utils.SetupModel import SetupModel, SetupPlayer
from pychess.Utils.TimeModel import TimeModel
from pychess.Utils.const import (
    NORMALCHESS,
    FISCHERRANDOMCHESS,
    VARIANTS_ODDS,
    VARIANTS_SHUFFLE,
    VARIANTS_OTHER,
    VARIANTS_OTHER_NONSTANDARD,
    VARIANTS_ASEAN,
    WHITE,
    BLACK,
    UNSUPPORTED,
    ARTIFICIAL,
    LOCAL,
    reprCord,
    reprFile,
    W_OO,
    W_OOO,
    B_OO,
    B_OOO,
    FAN_PIECES,
    reprSign,
    FEN_START,
    WAITING_TO_START,
)

from pychess.Utils.repr import localReprSign
from pychess.Utils.lutils.ldata import FILE
from pychess.Utils.lutils.LBoard import LBoard
from pychess.System import uistuff
from pychess.System.Log import log
from pychess.System.protoopen import splitUri
from pychess.System import conf
from pychess.System.prefix import getDataPrefix, isInstalled, addDataPrefix
from pychess.Players.engineNest import discoverer
from pychess.Players.Human import Human
from pychess.widgets import ImageMenu, mainwindow
from pychess.widgets.prompttext import getUserTextDialog
from pychess.widgets.BoardControl import BoardControl
from pychess.Savers import fen, pgn
from pychess.Savers.ChessFile import LoadingError
from pychess.Variants import variants
from pychess.Variants.normal import NormalBoard
from pychess.perspectives import perspective_manager
from pychess.perspectives.games import enddir
from pychess.Utils.eco import find_opening_fen

# ===============================================================================
# We init most dialog icons global to make them accessibly to the
# Background.Taskers so they have a similar look.
# ===============================================================================
big_time = get_pixbuf("glade/stock_alarm.svg")
big_people = get_pixbuf("glade/people48.png")
iwheels = get_pixbuf("glade/wheel.png")
ipeople = get_pixbuf("glade/people24.png")
inotebook = load_icon(24, "stock_notebook", "computer")

weather_icons = (
    "clear",
    "clear-night",
    "few-clouds",
    "few-clouds-night",
    "fog",
    "overcast",
    "severe-alert",
    "showers-scattered",
    "showers",
    "storm",
)
skillToIcon = {}
# Used by TaskerManager. Put here to help synchronization
skillToIconLarge = {}
for i, icon in enumerate(weather_icons, start=1):
    skillToIcon[2 * i - 1] = get_pixbuf("glade/16x16/weather-%s.png" % icon)
    skillToIcon[2 * i] = get_pixbuf("glade/16x16/weather-%s.png" % icon)
    skillToIconLarge[2 * i - 1] = get_pixbuf("glade/48x48/weather-%s.png" % icon)
    skillToIconLarge[2 * i] = get_pixbuf("glade/48x48/weather-%s.png" % icon)

playerItems = []
analyzerItems = []
allEngineItems = []


def createPlayerUIGlobals(discoverer):
    global playerItems
    global analyzerItems
    global allEngineItems

    playerItems = []
    analyzerItems = []
    allEngineItems = []

    for variantClass in variants.values():
        playerItems += [[(ipeople, _("Human Being"))]]
    for engine in discoverer.getEngines():
        name = engine["name"]
        c = discoverer.getCountry(engine)
        path = addDataPrefix("flags/%s.png" % c)

        if c and os.path.isfile(path):
            flag_icon = get_pixbuf(path)
        else:
            path = addDataPrefix("flags/unknown.png")
            flag_icon = get_pixbuf(path)

        allEngineItems.append((flag_icon, name))

        for variant in discoverer.getEngineVariants(engine):
            playerItems[variant] += [(flag_icon, name)]

        if discoverer.is_analyzer(engine):
            analyzerItems.append((flag_icon, name))


discoverer.connect("all_engines_discovered", createPlayerUIGlobals)

COPY, CLEAR, PASTE, INITIAL = 2, 3, 4, 5

# ===============================================================================
# GameInitializationMode is the super class of new game dialogs. Dialogs include
# the standard new game dialog, the load file dialog, the enter notation dialog
# and the setup position dialog.
# ===============================================================================


class _GameInitializationMode:
    @classmethod
    def _ensureReady(cls):
        if not hasattr(_GameInitializationMode, "superhasRunInit"):
            _GameInitializationMode._init()
            _GameInitializationMode.superhasRunInit = True
        if not hasattr(cls, "hasRunInit"):
            cls._init()
            cls.hasRunInit = True
        cls.widgets["newgamedialog"].resize(1, 1)

    @classmethod
    def _init(cls):
        cls.white = get_pixbuf("glade/white.png")
        cls.black = get_pixbuf("glade/black.png")

        cls.widgets = uistuff.GladeWidgets("newInOut.glade")
        cls.widgets["newgamedialog"].set_transient_for(mainwindow())

        def on_exchange_players(widget, button_event):
            white = cls.widgets["whitePlayerCombobox"].get_active()
            black = cls.widgets["blackPlayerCombobox"].get_active()
            whiteLevel = cls.widgets["skillSlider1"].get_value()
            blackLevel = cls.widgets["skillSlider2"].get_value()
            cls.widgets["whitePlayerCombobox"].set_active(black)
            cls.widgets["blackPlayerCombobox"].set_active(white)
            cls.widgets["skillSlider1"].set_value(blackLevel)
            cls.widgets["skillSlider2"].set_value(whiteLevel)

        cls.widgets["whitePlayerButton"].set_image(Gtk.Image.new_from_pixbuf(cls.white))
        cls.widgets["whitePlayerButton"].connect(
            "button-press-event", on_exchange_players
        )
        cls.widgets["blackPlayerButton"].set_image(Gtk.Image.new_from_pixbuf(cls.black))
        cls.widgets["blackPlayerButton"].connect(
            "button-press-event", on_exchange_players
        )

        uistuff.createCombo(
            cls.widgets["whitePlayerCombobox"], name="whitePlayerCombobox"
        )
        uistuff.createCombo(
            cls.widgets["blackPlayerCombobox"], name="blackPlayerCombobox"
        )

        cls.widgets["playersIcon"].set_from_pixbuf(big_people)
        cls.widgets["timeIcon"].set_from_pixbuf(big_time)

        def on_playerCombobox_changed(widget, skill_hbox, skill_level):
            position = widget.get_active()
            skill_hbox.props.visible = position > 0
            if position > 0:
                tree_iter = widget.get_active_iter()
                if tree_iter is not None:
                    engine_name = widget.get_model()[tree_iter][1]
                    engine = discoverer.getEngineByName(engine_name)
                    if engine:
                        pref_level = engine.get("level")
                        if pref_level:
                            skill_level.set_value(pref_level)

        cls.widgets["whitePlayerCombobox"].connect(
            "changed",
            on_playerCombobox_changed,
            cls.widgets["skillHbox1"],
            cls.widgets["skillSlider1"],
        )
        cls.widgets["blackPlayerCombobox"].connect(
            "changed",
            on_playerCombobox_changed,
            cls.widgets["skillHbox2"],
            cls.widgets["skillSlider2"],
        )
        cls.widgets["whitePlayerCombobox"].set_active(0)
        cls.widgets["blackPlayerCombobox"].set_active(1)

        def on_skill_changed(scale, image):
            image.set_from_pixbuf(skillToIcon[int(scale.get_value())])

        cls.widgets["skillSlider1"].connect(
            "value-changed", on_skill_changed, cls.widgets["skillIcon1"]
        )
        cls.widgets["skillSlider2"].connect(
            "value-changed", on_skill_changed, cls.widgets["skillIcon2"]
        )
        cls.widgets["skillSlider1"].set_value(3)
        cls.widgets["skillSlider2"].set_value(3)

        cls.__initTimeRadio(
            "ngblitz",
            cls.widgets["blitzRadio"],
            cls.widgets["configImageBlitz"],
            5,
            0,
            0,
        )
        cls.__initTimeRadio(
            "ngrapid",
            cls.widgets["rapidRadio"],
            cls.widgets["configImageRapid"],
            15,
            5,
            0,
        )
        cls.__initTimeRadio(
            "ngnormal",
            cls.widgets["normalRadio"],
            cls.widgets["configImageNormal"],
            45,
            15,
            0,
        )
        cls.__initTimeRadio(
            "ngclassical",
            cls.widgets["classicalRadio"],
            cls.widgets["configImageClassical"],
            3,
            0,
            40,
        )

        cls.__initVariantRadio(
            "ngvariant1",
            cls.widgets["playVariant1Radio"],
            cls.widgets["configImageVariant1"],
        )
        cls.__initVariantRadio(
            "ngvariant2",
            cls.widgets["playVariant2Radio"],
            cls.widgets["configImageVariant2"],
        )

        def updateCombos(*args):
            if cls.widgets["playNormalRadio"].get_active():
                variant = NORMALCHESS
            elif cls.widgets["playVariant1Radio"].get_active():
                variant = conf.get("ngvariant1")
            else:
                variant = conf.get("ngvariant2")
            variant1 = conf.get("ngvariant1")
            cls.widgets["playVariant1Radio"].set_tooltip_text(
                variants[variant1].__desc__
            )
            variant2 = conf.get("ngvariant2")
            cls.widgets["playVariant2Radio"].set_tooltip_text(
                variants[variant2].__desc__
            )
            data = [(item[0], item[1]) for item in playerItems[variant]]
            uistuff.updateCombo(cls.widgets["blackPlayerCombobox"], data)
            uistuff.updateCombo(cls.widgets["whitePlayerCombobox"], data)

        discoverer.connect_after("all_engines_discovered", updateCombos)
        updateCombos(discoverer)

        conf.notify_add("ngvariant1", updateCombos)
        conf.notify_add("ngvariant2", updateCombos)
        cls.widgets["playNormalRadio"].connect("toggled", updateCombos)
        cls.widgets["playNormalRadio"].set_tooltip_text(variants[NORMALCHESS].__desc__)
        cls.widgets["playVariant1Radio"].connect("toggled", updateCombos)
        variant1 = conf.get("ngvariant1")
        cls.widgets["playVariant1Radio"].set_tooltip_text(variants[variant1].__desc__)
        cls.widgets["playVariant2Radio"].connect("toggled", updateCombos)
        variant2 = conf.get("ngvariant2")
        cls.widgets["playVariant2Radio"].set_tooltip_text(variants[variant2].__desc__)

        # The "variant" has to come before players, because the engine positions
        # in the user comboboxes can be different in different variants
        for key in (
            "whitePlayerCombobox",
            "blackPlayerCombobox",
            "skillSlider1",
            "skillSlider2",
            "notimeRadio",
            "blitzRadio",
            "rapidRadio",
            "normalRadio",
            "classicalRadio",
            "playNormalRadio",
            "playVariant1Radio",
            "playVariant2Radio",
        ):
            uistuff.keep(cls.widgets[key], key)

        # We don't want the dialog to deallocate when closed. Rather we hide
        # it on respond
        cls.widgets["newgamedialog"].connect("delete_event", lambda *a: True)

    @classmethod
    def __initTimeRadio(cls, id, radiobutton, configImage, defmin, defgain, defmoves):
        minSpin = Gtk.SpinButton()
        minSpin.set_adjustment(Gtk.Adjustment(1, 0, 240, 1))
        setattr(cls, "%s_min" % id, minSpin)
        uistuff.keep(minSpin, "%s min" % id)
        movesSpin = Gtk.SpinButton()
        movesSpin.set_adjustment(Gtk.Adjustment(0, 0, 60, 20))
        setattr(cls, "%s_moves" % id, movesSpin)
        uistuff.keep(movesSpin, "%s moves" % id)
        gainSpin = Gtk.SpinButton()
        gainSpin.set_adjustment(Gtk.Adjustment(0, -60, 60, 1))
        setattr(cls, "%s_gain" % id, gainSpin)
        uistuff.keep(gainSpin, "%s gain" % id)

        table = Gtk.Table(2, 2)
        table.props.row_spacing = 3
        table.props.column_spacing = 12
        label = Gtk.Label(label=_("Minutes:"))
        label.props.xalign = 0
        table.attach(label, 0, 1, 0, 1)
        table.attach(minSpin, 1, 2, 0, 1)
        label = Gtk.Label(label=_("Moves:") if defmoves > 0 else _("Gain:"))
        label.props.xalign = 0
        table.attach(label, 0, 1, 1, 2)
        if defmoves > 0:
            table.attach(movesSpin, 1, 2, 1, 2)
        else:
            table.attach(gainSpin, 1, 2, 1, 2)
        alignment = Gtk.Alignment.new(1, 1, 1, 1)
        alignment.set_padding(6, 6, 12, 12)
        alignment.add(table)
        ImageMenu.switchWithImage(configImage, alignment)

        def updateString(spin):
            # Elements of the clock
            minutes = minSpin.get_value_as_int()
            gain = gainSpin.get_value_as_int()
            moves = movesSpin.get_value_as_int()

            # Duration of the game
            def calculate_duration(ref_moves):
                if moves > 0:
                    return int(2 * minutes * ref_moves / moves)
                else:
                    return max(0, int(2 * minutes + ref_moves * gain / 30))

            duration_20 = calculate_duration(20)
            duration_40 = calculate_duration(40)
            duration_60 = calculate_duration(60)

            # Determination of the caption
            def get_game_name():
                """https://www.fide.com/fide/handbook.html?id=171&view=article"""
                if defmoves > 0:
                    return _("Classical")
                if duration_60 <= 20:  # 10 minutes per player
                    return _("Blitz")
                if duration_60 < 120:  # 60 minutes per player
                    return _("Rapid")
                return _("Normal")

            if moves > 0:
                radiobutton.set_label(
                    _("%(name)s %(minutes)d min / %(moves)d moves %(duration)s")
                    % {
                        "name": get_game_name(),
                        "minutes": minutes,
                        "moves": moves,
                        "duration": ("(%d')" % duration_40) if duration_40 > 0 else "",
                    }
                )
            elif gain != 0:
                radiobutton.set_label(
                    _(
                        "%(name)s %(minutes)d min %(sign)s %(gain)d sec/move %(duration)s"
                    )
                    % {
                        "name": get_game_name(),
                        "minutes": minutes,
                        "sign": "+" if gain > 0 else "–",
                        "gain": abs(gain),
                        "duration": ("(%d')" % duration_40) if duration_40 > 0 else "",
                    }
                )
            else:
                radiobutton.set_label(
                    _("%(name)s %(minutes)d min %(duration)s")
                    % {
                        "name": get_game_name(),
                        "minutes": minutes,
                        "duration": ("(%d')" % duration_40) if duration_40 > 0 else "",
                    }
                )

            # Determination of the tooltip
            if duration_20 > 0 and duration_60 > 0 and duration_20 != duration_60:
                radiobutton.set_tooltip_text(
                    _("Estimated duration : %(min)d - %(max)d minutes")
                    % (
                        {
                            "min": min(duration_20, duration_60),
                            "max": max(duration_20, duration_60),
                        }
                    )
                )
            else:
                radiobutton.set_tooltip_text("")

        minSpin.connect("value-changed", updateString)
        movesSpin.connect("value-changed", updateString)
        gainSpin.connect("value-changed", updateString)
        updateString(None)

    @classmethod
    def __initVariantRadio(cls, confid, radiobutton, configImage):
        model = Gtk.TreeStore(str)
        treeview = Gtk.TreeView(model)
        treeview.set_headers_visible(False)
        treeview.append_column(Gtk.TreeViewColumn(None, Gtk.CellRendererText(), text=0))
        alignment = Gtk.Alignment.new(1, 1, 1, 1)
        alignment.set_padding(6, 6, 12, 12)
        alignment.add(treeview)
        ImageMenu.switchWithImage(configImage, alignment)

        groupNames = {
            VARIANTS_ODDS: _("Odds"),
            VARIANTS_SHUFFLE: _("Shuffle"),
            VARIANTS_OTHER: _("Other (standard rules)"),
            VARIANTS_OTHER_NONSTANDARD: _("Other (non standard rules)"),
            VARIANTS_ASEAN: _("Asian variants"),
        }

        specialVariants = [
            v
            for v in variants.values()
            if v != NormalBoard and v.variant not in UNSUPPORTED
        ]
        specialVariants = sorted(specialVariants, key=attrgetter("variant_group"))
        groups = groupby(specialVariants, attrgetter("variant_group"))
        pathToVariant = {}
        variantToPath = {}
        for i, (id, group) in enumerate(groups):
            iter = model.append(None, (groupNames[id],))
            for variant in group:
                subiter = model.append(iter, (variant.name,))
                path = model.get_path(subiter)
                pathToVariant[path.to_string()] = variant.variant
                variantToPath[variant.variant] = path.to_string()
            treeview.expand_row(Gtk.TreePath(i), True)

        selection = treeview.get_selection()
        selection.set_mode(Gtk.SelectionMode.BROWSE)

        def selfunc(selection, store, path, path_selected, data):
            return path.get_depth() > 1

        selection.set_select_function(selfunc, None)
        variant = conf.get(confid)
        if variant in variantToPath:
            selection.select_path(variantToPath[variant])

        def callback(selection):
            model, iter = selection.get_selected()
            if iter:
                radiobutton.set_label("%s" % model.get(iter, 0) + _(" chess"))
                path = model.get_path(iter)
                variant = pathToVariant[path.to_string()]
                conf.set(confid, variant)

        selection.connect("changed", callback)
        callback(selection)

    @classmethod
    def _generalRun(cls, callback, validate):
        def onResponse(dialog, response):
            if response == COPY:
                clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
                clipboard.set_text(cls.get_fen(), -1)
                # print("put clipboard:", clipboard.wait_for_text())
                return
            elif response == CLEAR:
                cls.board_control.emit("action", "SETUP", None, True)
                cls.ini_widgets(True)
                # print("clear")
                return
            elif response == PASTE:
                clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
                text = clipboard.wait_for_text()
                # print("got clipboard:", text)
                if text is None or len(text.split()) < 2:
                    return
                try:
                    lboard = cls.setupmodel.variant(setup=text).board
                    cls.ini_widgets(lboard.asFen())
                    cls.board_control.emit("action", "SETUP", None, text)
                except SyntaxError as e:
                    d = Gtk.MessageDialog(
                        mainwindow(),
                        type=Gtk.MessageType.WARNING,
                        buttons=Gtk.ButtonsType.OK,
                        message_format=e.args[0],
                    )
                    if len(e.args) > 1:
                        d.format_secondary_text(e.args[1])
                    d.connect("response", lambda d, a: d.hide())
                    d.show()
                return
            elif response == INITIAL:
                if os.environ.get(
                    "PYCHESS_UNITTEST"
                ):  # The modal dialog cannot be automated in the background
                    ecofen = FEN_START
                else:
                    ecoterms = getUserTextDialog(
                        dialog,
                        _("Start position from the opening book"),
                        _("Type an ECO code or a keyword (with wildcard *):"),
                    )
                    ecofen = find_opening_fen(ecoterms)
                    if ecofen is None:
                        if ecoterms != "":
                            dlgwin = Gtk.MessageDialog(
                                mainwindow(),
                                type=Gtk.MessageType.QUESTION,
                                buttons=Gtk.ButtonsType.YES_NO,
                            )
                            dlgwin.set_markup(
                                _(
                                    "No position was found. Do you want to use the default setup ?"
                                )
                            )
                            dlgrep = dlgwin.run()
                            dlgwin.destroy()
                            if dlgrep != Gtk.ResponseType.YES:
                                return
                        ecofen = FEN_START
                lboard = cls.setupmodel.variant(setup=ecofen).board
                cls.ini_widgets(lboard.asFen())
                cls.board_control.emit("action", "SETUP", None, ecofen)
                return
            elif response != Gtk.ResponseType.OK:
                cls.widgets["newgamedialog"].hide()
                cls.widgets["newgamedialog"].disconnect(handlerId)
                return

            if hasattr(cls, "board_control"):
                cls.board_control.emit("action", "CLOSE", None, None)

            # Find variant
            if cls.widgets["playNormalRadio"].get_active():
                variant_index = NORMALCHESS
            elif cls.widgets["playVariant1Radio"].get_active():
                variant_index = conf.get("ngvariant1")
            else:
                variant_index = conf.get("ngvariant2")
            variant = variants[variant_index]

            # Find time
            if cls.widgets["notimeRadio"].get_active():
                secs = 0
                incr = 0
                moves = 0
            elif cls.widgets["blitzRadio"].get_active():
                secs = cls.ngblitz_min.get_value_as_int() * 60
                incr = cls.ngblitz_gain.get_value_as_int()
                moves = 0
            elif cls.widgets["rapidRadio"].get_active():
                secs = cls.ngrapid_min.get_value_as_int() * 60
                incr = cls.ngrapid_gain.get_value_as_int()
                moves = 0
            elif cls.widgets["normalRadio"].get_active():
                secs = cls.ngnormal_min.get_value_as_int() * 60
                incr = cls.ngnormal_gain.get_value_as_int()
                moves = 0
            elif cls.widgets["classicalRadio"].get_active():
                secs = cls.ngclassical_min.get_value_as_int() * 60
                incr = 0
                moves = cls.ngclassical_moves.get_value_as_int()

            # Find players
            player0combo = cls.widgets["whitePlayerCombobox"]
            player0 = player0combo.get_active()

            tree_iter = player0combo.get_active_iter()
            if tree_iter is not None:
                model = player0combo.get_model()
                name0 = model[tree_iter][1]

            diffi0 = int(cls.widgets["skillSlider1"].get_value())

            player1combo = cls.widgets["blackPlayerCombobox"]
            player1 = player1combo.get_active()

            tree_iter = player1combo.get_active_iter()
            if tree_iter is not None:
                model = player1combo.get_model()
                name1 = model[tree_iter][1]

            diffi1 = int(cls.widgets["skillSlider2"].get_value())

            # Prepare players
            playertups = []
            for i, playerno, name, diffi, color in (
                (0, player0, name0, diffi0, WHITE),
                (1, player1, name1, diffi1, BLACK),
            ):
                if playerno > 0:
                    engine = discoverer.getEngineByName(name)
                    playertups.append(
                        (
                            ARTIFICIAL,
                            discoverer.initPlayerEngine,
                            [engine, color, diffi, variant, secs, incr, moves],
                            name,
                        )
                    )
                else:
                    if not playertups or playertups[0][0] != LOCAL:
                        name = conf.get("firstName")
                    else:
                        name = conf.get("secondName")
                    playertups.append((LOCAL, Human, (color, name), name))

            # Set forcePonderOff initPlayerEngine param True in engine-engine games
            if playertups[0][0] == ARTIFICIAL and playertups[1][0] == ARTIFICIAL:
                playertups[0][2].append(True)
                playertups[1][2].append(True)

            timemodel = TimeModel(secs, incr, moves=moves)
            gamemodel = GameModel(timemodel, variant)

            if not validate(gamemodel):
                return
            else:
                cls.widgets["newgamedialog"].hide()
                cls.widgets["newgamedialog"].disconnect(handlerId)
                callback(gamemodel, playertups[0], playertups[1])

        handlerId = cls.widgets["newgamedialog"].connect("response", onResponse)
        cls.widgets["newgamedialog"].show()

    @classmethod
    def _hideOthers(cls):
        for extension in (
            "loadsidepanel",
            "enterGameNotationSidePanel",
            "setupPositionSidePanel",
        ):
            cls.widgets[extension].hide()
        for button in ("copy_button", "clear_button", "paste_button", "initial_button"):
            cls.widgets[button].hide()


# ###############################################################################
# NewGameMode                                                                  #
# ###############################################################################


class NewGameMode(_GameInitializationMode):
    @classmethod
    def _init(cls):
        # We have to override this, so the GameInitializationMode init method
        # isn't called twice
        pass

    @classmethod
    def run(cls):
        cls._ensureReady()
        if cls.widgets["newgamedialog"].props.visible:
            cls.widgets["newgamedialog"].present()
            return

        def _validate(gamemodel):
            return True

        cls._hideOthers()
        cls.widgets["newgamedialog"].set_title(_("New Game"))

        def _callback(gamemodel, p0, p1):
            perspective = perspective_manager.get_perspective("games")
            asyncio.create_task(perspective.generalStart(gamemodel, p0, p1))

        cls._generalRun(_callback, _validate)


# ###############################################################################
# SetupPositionExtension                                                       #
# ###############################################################################


class SetupPositionExtension(_GameInitializationMode):
    board_control = None

    @classmethod
    def _init(cls):
        def callback(widget, allocation):
            cls.widgets["setupPositionFrame"].set_size_request(
                523, allocation.height - 4
            )

        cls.widgets["setupPositionSidePanel"].connect_after("size-allocate", callback)

        cls.castl = set()

        cls.white = Gtk.Image.new_from_pixbuf(cls.white)
        cls.black = Gtk.Image.new_from_pixbuf(cls.black)
        cls.widgets["side_button"].set_image(cls.white)

        cls.widgets["side_button"].connect("toggled", cls.side_button_toggled)
        cls.widgets["rotate_button"].connect(
            "button-press-event", cls.rotate_button_pressed
        )
        cls.widgets["moveno_spin"].connect("value-changed", cls.moveno_spin_changed)
        cls.widgets["fifty_spin"].connect("value-changed", cls.fifty_spin_changed)

        cls.widgets["woo"].connect("toggled", cls.castl_toggled, W_OO)
        cls.widgets["wooo"].connect("toggled", cls.castl_toggled, W_OOO)
        cls.widgets["boo"].connect("toggled", cls.castl_toggled, B_OO)
        cls.widgets["booo"].connect("toggled", cls.castl_toggled, B_OOO)

        ep_store = Gtk.ListStore(str)
        ep_store.append(["-"])
        for f in reprFile:
            ep_store.append([f])
        epcombo = cls.widgets["ep_combo"]
        epcombo.set_name("ep_combo")
        epcombo.set_model(ep_store)
        renderer_text = Gtk.CellRendererText()
        cls.widgets["ep_combo"].pack_start(renderer_text, True)
        cls.widgets["ep_combo"].add_attribute(renderer_text, "text", 0)
        cls.widgets["ep_combo"].set_active(0)
        cls.widgets["ep_combo"].connect("changed", cls.ep_combo_changed)

        cls.widgets["playNormalRadio"].connect("toggled", cls.fen_changed)
        cls.widgets["playVariant1Radio"].connect("toggled", cls.fen_changed)
        cls.widgets["playVariant2Radio"].connect("toggled", cls.fen_changed)

    @classmethod
    def side_button_toggled(cls, button):
        if button.get_active():
            button.set_image(cls.black)
        else:
            button.set_image(cls.white)
        cls.fen_changed()

    @classmethod
    def rotate_button_pressed(self, widget, button_event):
        view = self.board_control.view
        if view.rotation:
            view.rotation = 0
        else:
            view.rotation = pi

    @classmethod
    def fen_changed(cls, *args):
        cls.widgets["fen_entry"].set_text(cls.get_fen())

    @classmethod
    def game_changed(cls, model, ply):
        GLib.idle_add(cls.fen_changed)

    @classmethod
    def ep_combo_changed(cls, combo):
        cls.fen_changed()

    @classmethod
    def moveno_spin_changed(cls, spin):
        cls.fen_changed()

    @classmethod
    def fifty_spin_changed(cls, spin):
        cls.fen_changed()

    @classmethod
    def castl_toggled(cls, button, castl):
        lboard = cls.setupmodel.boards[-1].board
        # TODO: this doesn't work at all
        if lboard.variant == FISCHERRANDOMCHESS:
            if castl == W_OO:
                cast_letter = reprCord[lboard.ini_rooks[0][1]][0].upper()
            elif castl == W_OOO:
                cast_letter = reprCord[lboard.ini_rooks[0][0]][0].upper()
            elif castl == B_OO:
                cast_letter = reprCord[lboard.ini_rooks[1][1]][0]
            elif castl == B_OOO:
                cast_letter = reprCord[lboard.ini_rooks[1][0]][0]
        else:
            if castl == W_OO:
                cast_letter = "K"
            elif castl == W_OOO:
                cast_letter = "Q"
            elif castl == B_OO:
                cast_letter = "k"
            elif castl == B_OOO:
                cast_letter = "q"

        if button.get_active():
            cls.castl.add(cast_letter)
        else:
            cls.castl.discard(cast_letter)
        cls.fen_changed()

    @classmethod
    def get_fen(cls):
        # Find variant
        if cls.widgets["playNormalRadio"].get_active():
            variant_index = NORMALCHESS
        elif cls.widgets["playVariant1Radio"].get_active():
            variant_index = conf.get("ngvariant1")
        else:
            variant_index = conf.get("ngvariant2")
        variant = variants[variant_index]

        pieces = cls.setupmodel.boards[-1].as_fen(variant.variant)

        side = "b" if cls.widgets["side_button"].get_active() else "w"
        castl = "".join(sorted(cls.castl)) if cls.castl else "-"

        ep = "-"
        rank = "3" if side == "b" else "6"
        tree_iter = cls.widgets["ep_combo"].get_active_iter()
        if tree_iter is not None:
            model = cls.widgets["ep_combo"].get_model()
            ep = model[tree_iter][0]
        ep = ep if ep == "-" else ep + rank

        fifty = cls.widgets["fifty_spin"].get_value_as_int()
        moveno = cls.widgets["moveno_spin"].get_value_as_int()

        parts = (pieces, side, castl, ep, fifty, moveno)
        return "%s %s %s %s %s %s" % parts

    @classmethod
    def ini_widgets(cls, setup, lboard=None):
        if lboard is None:
            lboard = cls.setupmodel.variant(setup=setup).board
        cls.widgets["side_button"].set_active(False if lboard.color == WHITE else True)
        cls.widgets["fifty_spin"].set_value(lboard.fifty)
        cls.widgets["moveno_spin"].set_value(lboard.plyCount // 2 + 1)
        ep = lboard.enpassant
        cls.widgets["ep_combo"].set_active(0 if ep is None else FILE(ep) + 1)
        cls.widgets["woo"].set_active(lboard.castling & W_OO)
        cls.widgets["wooo"].set_active(lboard.castling & W_OOO)
        cls.widgets["boo"].set_active(lboard.castling & B_OO)
        cls.widgets["booo"].set_active(lboard.castling & B_OOO)

    @classmethod
    def run(cls, fenstr, variant):
        cls._ensureReady()
        if cls.widgets["newgamedialog"].props.visible:
            cls.widgets["newgamedialog"].present()
            return

        cls._hideOthers()
        for button in ("copy_button", "clear_button", "paste_button", "initial_button"):
            cls.widgets[button].show()
        cls.widgets["newgamedialog"].set_title(_("Setup Position"))
        cls.widgets["setupPositionSidePanel"].show()

        cls.setupmodel = SetupModel()
        cls.board_control = BoardControl(cls.setupmodel, {}, setup_position=True)
        cls.setupmodel.curplayer = SetupPlayer(cls.board_control)
        cls.setupmodel.connect("game_changed", cls.game_changed)

        child = cls.widgets["setupBoardDock"].get_child()
        if child is not None:
            cls.widgets["setupBoardDock"].remove(child)
        cls.widgets["setupBoardDock"].add(cls.board_control)
        cls.board_control.show_all()
        if fenstr is not None:
            lboard = LBoard(variant)
            lboard.applyFen(fenstr)
            cls.setupmodel.boards = [
                cls.setupmodel.variant(setup=fenstr, lboard=lboard)
            ]
            cls.setupmodel.variations = [cls.setupmodel.boards]
            cls.ini_widgets(fenstr, lboard)
        else:
            fenstr = cls.get_fen()
            cls.ini_widgets(True)
        cls.widgets["fen_entry"].set_text(fenstr)

        cls.setupmodel.start()
        cls.board_control.emit("action", "SETUP", None, fenstr)

        def _validate(gamemodel):
            try:
                fenstr = cls.get_fen()
                cls.setupmodel.variant(setup=fenstr)
                return True
            except (AssertionError, LoadingError, SyntaxError) as e:
                d = Gtk.MessageDialog(
                    mainwindow(),
                    type=Gtk.MessageType.WARNING,
                    buttons=Gtk.ButtonsType.OK,
                    message_format=e.args[0],
                )
                if len(e.args) > 1:
                    d.format_secondary_text(e.args[1])
                d.connect("response", lambda d, a: d.hide())
                d.show()
                return False

        def _callback(gamemodel, p0, p1):
            text = cls.get_fen()
            perspective = perspective_manager.get_perspective("games")
            asyncio.create_task(
                perspective.generalStart(
                    gamemodel, p0, p1, (StringIO(text), fen, 0, -1)
                )
            )

        cls._generalRun(_callback, _validate)


# ###############################################################################
# EnterNotationExtension                                                       #
# ###############################################################################


class EnterNotationExtension(_GameInitializationMode):
    @classmethod
    def _init(cls):
        def callback(widget, allocation):
            cls.widgets["enterGameNotationFrame"].set_size_request(
                223, allocation.height - 4
            )

        cls.widgets["enterGameNotationSidePanel"].connect_after(
            "size-allocate", callback
        )

        flags = []
        if isInstalled():
            path = gettext.find("pychess")
        else:
            path = gettext.find("pychess", localedir=addDataPrefix("lang"))
        if path:
            default_locale = locale.getdefaultlocale()
            if default_locale[0] is not None:
                loc = locale.getdefaultlocale()[0][-2:].lower()
                flags.append(addDataPrefix("flags/%s.png" % loc))

        flags.append(addDataPrefix("flags/us.png"))

        cls.ib = ImageButton(flags)
        cls.widgets["imageButtonDock"].add(cls.ib)
        cls.ib.show()

        cls.sourcebuffer = GtkSource.Buffer()
        sourceview = GtkSource.View.new_with_buffer(cls.sourcebuffer)
        sourceview.set_tooltip_text(_("Type or paste PGN game or FEN positions here"))
        cls.widgets["scrolledwindow6"].add(sourceview)
        sourceview.show()

        # Pgn format does not allow tabulator
        sourceview.set_insert_spaces_instead_of_tabs(True)
        sourceview.set_wrap_mode(Gtk.WrapMode.WORD)

        man = GtkSource.LanguageManager()
        # Init new version
        if hasattr(man.props, "search_path"):
            try:
                path = os.path.join(getDataPrefix(), "gtksourceview-4/language-specs")
                man.props.search_path = man.props.search_path + [path]
                if "pgn" in man.get_language_ids():
                    lang = man.get_language("pgn")
                    cls.sourcebuffer.set_language(lang)
                else:
                    log.warning("Unable to load pgn syntax-highlighting.")
                cls.sourcebuffer.set_highlight_syntax(True)
            except NotImplementedError:
                # Python 2.7.3 in Ubuntu 12.04
                log.warning("Unable to load pgn syntax-highlighting.")
        # Init old version
        else:
            os.environ["XDG_DATA_DIRS"] = getDataPrefix() + ":/usr/share/"
            man = GtkSource.LanguageManager()
            for lang in man.get_available_languages():
                if lang.get_name() == "PGN":
                    cls.sourcebuffer.set_language(lang)
                    break
            else:
                log.warning("Unable to load pgn syntax-highlighting.")
            cls.sourcebuffer.set_highlight(True)

    @classmethod
    def run(cls):
        cls._ensureReady()
        if cls.widgets["newgamedialog"].props.visible:
            cls.widgets["newgamedialog"].present()
            return

        cls._hideOthers()
        cls.widgets["newgamedialog"].set_title(_("Enter Game"))
        cls.widgets["enterGameNotationSidePanel"].show()

        def _get_text():
            text = cls.sourcebuffer.get_text(
                cls.sourcebuffer.get_start_iter(),
                cls.sourcebuffer.get_end_iter(),
                False,
            )

            # Test if the ImageButton has two layers and is set on the local language
            if len(cls.ib.surfaces) == 2 and cls.ib.current == 0:
                # 2 step used to avoid backtranslating
                # (local and english piece letters can overlap)
                for i, sign in enumerate(localReprSign[1:]):
                    if sign.strip():
                        text = text.replace(sign, FAN_PIECES[0][i + 1])
                for i, sign in enumerate(FAN_PIECES[0][1:7]):
                    text = text.replace(sign, reprSign[i + 1])
                text = str(text)

            # First we try if it's just a FEN string
            parts_no = len(text.split())
            if text.strip() == "":
                text = FEN_START
                loadType = fen
            elif parts_no > 0 and text.split()[0].count("/") == 7:
                loadType = fen
            else:
                # patch default human player names on demand...
                player0combo = cls.widgets["whitePlayerCombobox"]
                player0 = player0combo.get_active()
                if player0 == 0 and '[White "' not in text:
                    name = '[White "%s"]' % conf.get("firstName")
                    text = f"{name}\n{text}"

                player1combo = cls.widgets["blackPlayerCombobox"]
                player1 = player1combo.get_active()
                if player1 == 0 and '[Black "' not in text:
                    name = '[Black "%s"]' % conf.get("secondName")
                    text = f"{name}\n{text}"

                loadType = pgn

            return text, loadType

        def _validate(gamemodel):
            try:
                text, loadType = _get_text()
                chessfile = loadType.load(StringIO(text))
                chessfile.loadToModel(chessfile.games[0], -1, model=gamemodel)
                gamemodel.status = WAITING_TO_START
                return True
            except LoadingError as e:
                d = Gtk.MessageDialog(
                    mainwindow(),
                    type=Gtk.MessageType.WARNING,
                    buttons=Gtk.ButtonsType.OK,
                    message_format=e.args[0],
                )
                d.format_secondary_text(e.args[1])
                d.connect("response", lambda d, a: d.hide())
                d.show()
                return False

        def _callback(gamemodel, p0, p1):
            text, loadType = _get_text()
            perspective = perspective_manager.get_perspective("games")
            asyncio.create_task(
                perspective.generalStart(
                    gamemodel, p0, p1, (StringIO(text), loadType, 0, -1)
                )
            )

        cls._generalRun(_callback, _validate)


class ImageButton(Gtk.Button):
    def __init__(self, image_paths):
        GObject.GObject.__init__(self)

        self.surfaces = [Gtk.Image().new_from_file(path) for path in image_paths]
        self.current = 0

        self.image = self.surfaces[self.current]
        self.image.show()
        self.add(self.image)

        self.connect("clicked", self.on_clicked)

    def on_clicked(self, button):
        self.current = (self.current + 1) % len(self.surfaces)
        self.remove(self.image)
        self.image = self.surfaces[self.current]
        self.image.show()
        self.add(self.image)


def createRematch(gamemodel):
    """If gamemodel contains only LOCAL or ARTIFICIAL players, this starts a
    new game, based on the info in gamemodel"""

    if gamemodel.timed:
        secs = gamemodel.timemodel.intervals[0][WHITE]
        gain = gamemodel.timemodel.gain
        moves = gamemodel.timemodel.moves
    else:
        secs = 0
        gain = 0
        moves = 0
    newgamemodel = GameModel(TimeModel(secs, gain), variant=gamemodel.variant)

    wp = gamemodel.players[WHITE]
    bp = gamemodel.players[BLACK]

    if wp.__type__ == LOCAL:
        player1tup = (wp.__type__, wp.__class__, (BLACK, repr(wp)), repr(wp))
        if bp.__type__ == LOCAL:
            player0tup = (bp.__type__, bp.__class__, (WHITE, repr(wp)), repr(bp))
        else:
            engine = discoverer.getEngineByMd5(bp.md5)
            player0tup = (
                ARTIFICIAL,
                discoverer.initPlayerEngine,
                (engine, WHITE, bp.strength, gamemodel.variant, secs, gain, moves),
                repr(bp),
            )
    else:
        player0tup = (bp.__type__, bp.__class__, (WHITE, repr(bp)), repr(bp))
        engine = discoverer.getEngineByMd5(wp.md5)
        player1tup = (
            ARTIFICIAL,
            discoverer.initPlayerEngine,
            (engine, BLACK, wp.strength, gamemodel.variant, secs, gain),
            repr(wp),
        )

    perspective = perspective_manager.get_perspective("games")
    asyncio.create_task(perspective.generalStart(newgamemodel, player0tup, player1tup))


def loadFileAndRun(uri):
    if uri in [None, ""]:
        return False
    parts = splitUri(uri)
    uri = parts[1] if len(parts) == 2 else parts[0]
    loader = enddir[uri[uri.rfind(".") + 1 :]]
    timemodel = TimeModel(0, 0)
    gamemodel = GameModel(timemodel)
    white_name = _("White")
    black_name = _("Black")
    p0 = (LOCAL, Human, (WHITE, white_name), white_name)
    p1 = (LOCAL, Human, (BLACK, black_name), black_name)
    perspective = perspective_manager.get_perspective("games")
    asyncio.create_task(
        perspective.generalStart(gamemodel, p0, p1, (uri, loader, 0, -1))
    )
    return True


def loadPgnAndRun(data):
    if data in [None, ""]:
        return False
    perspective = perspective_manager.get_perspective("games")
    p0 = (LOCAL, Human, (WHITE, _("White")), _("White"))
    p1 = (LOCAL, Human, (BLACK, _("Black")), _("Black"))
    asyncio.create_task(
        perspective.generalStart(
            GameModel(), p0, p1, (StringIO(data), enddir["pgn"], 0, -1)
        )
    )
    return True



from pychess.System import conf
from pychess.Utils.const import FAN_PIECES, BLACK, WHITE

__label__ = _("Text Diagram")
__ending__ = "txt"
__append__ = True


def save(file, model, position=None, flip=False):
    """Export the current position into a .txt file using unicode chars"""

    data = model.boards[position].data[:]

    show_cords = conf.get("showCords")
    cords_side = "12345678" if flip else "87654321"
    cords_bottom = "hgfedcba" if flip else "abcdefgh"

    board = ""
    for j, row in enumerate(data if flip else reversed(data)):
        for i in range(8):
            piece = row.get(i)
            if piece is not None:
                if piece.color == BLACK:
                    piece_fan = FAN_PIECES[BLACK][piece.piece]
                else:
                    piece_fan = FAN_PIECES[WHITE][piece.piece]
                board += piece_fan
            else:
                board += "."
        if show_cords:
            board += cords_side[j]
        board += "\n"

    if show_cords:
        board += cords_bottom + "\n"

    print(board, file=file)

    file.close()
